import json
import os
import pickle
from functools import wraps
from typing import Callable, cast
from unittest.mock import Mock, patch

import boto3
import pytest
from kubernetes import client

import redun.executors.k8s
from redun import File, job_array, task
from redun.cli import RedunClient, import_script
from redun.config import Config
from redun.executors.code_packaging import create_tar, package_code
from redun.executors.command import REDUN_REQUIRED_VERSION
from redun.executors.k8s import (
    K8SExecutor,
    get_hash_from_job_name,
    get_k8s_job_name,
    submit_task,
)
from redun.executors.k8s_utils import DEFAULT_JOB_PREFIX, K8SClient, create_job_object
from redun.executors.scratch import (
    SCRATCH_ERROR,
    SCRATCH_HASHES,
    SCRATCH_INPUT,
    SCRATCH_OUTPUT,
    get_array_scratch_file,
    get_job_scratch_file,
)
from redun.file import Dir
from redun.scheduler import Job, Scheduler, Traceback
from redun.tests.utils import mock_s3, mock_scheduler, use_tempdir, wait_until
from redun.utils import pickle_dumps


def mock_k8s(func: Callable) -> Callable:
    """
    Mock k8s API clients during tests.
    """

    @wraps(func)
    def wrapped(*args, **kwargs):
        with patch.object(K8SClient, "core"), patch.object(K8SClient, "batch"), patch.object(
            K8SClient, "version", return_value=(1, 23)
        ):
            return func(*args, **kwargs)

    return wrapped


@pytest.mark.parametrize("array,suffix", [(False, ""), (True, "-array")])
def test_get_hash_from_job_name(array: bool, suffix: str) -> None:
    """
    Returns the Job hash from a k8s job name.
    """
    prefix = "my-job-prefix"
    job_hash = "c000d7f9b6275c58aff9d5466f6a1174e99195ca"
    job_name = get_k8s_job_name(prefix, job_hash, array=array)
    assert job_name.startswith(prefix)
    assert job_name.endswith(suffix)

    job_hash2 = get_hash_from_job_name(job_name)
    assert job_hash2 == job_hash


def mock_executor(scheduler: Scheduler, code_package: bool = False) -> K8SExecutor:
    """
    Returns an K8SExecutor with AWS API mocks.
    """
    image = "my-image"
    scratch_prefix = "s3://example-bucket/redun/"

    # Setup executor.
    config = Config(
        {
            "k8s": {
                "image": image,
                "scratch": scratch_prefix,
                "job_monitor_interval": 0.05,
                "job_stale_time": 0.01,
                "code_package": code_package,
            }
        }
    )
    executor = K8SExecutor("k8s", scheduler, config["k8s"])

    executor.get_jobs = Mock()  # type: ignore
    executor.get_jobs.return_value = []

    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")

    return executor


# skipped batch tags tests here


@mock_k8s
def test_executor_config(scheduler: Scheduler) -> None:
    """
    Executor should be able to parse its config.
    """
    # Setup executor.
    config = Config(
        {
            "k8s": {
                "image": "image",
                "scratch": "scratch_prefix",
                "code_includes": "*.txt",
            }
        }
    )
    executor = K8SExecutor("k8s", scheduler, config["k8s"])

    assert executor.image == "image"
    assert executor.scratch_prefix == "scratch_prefix"
    assert isinstance(executor.code_package, dict)
    assert executor.code_package["includes"] == ["*.txt"]


@task()
def task1(x: int):
    return x + 10


@task(load_module="custom.module")
def task1_custom_module(x):
    return x + 10


@use_tempdir
@mock_s3
@mock_k8s
@patch("redun.executors.k8s.k8s_submit")
@pytest.mark.parametrize(
    "custom_module, expected_load_module, a_task",
    [
        (None, "redun.tests.test_k8s", task1),
        ("custom.module", "custom.module", task1_custom_module),
    ],
)
def test_submit_task(k8s_submit_mock, custom_module, expected_load_module, a_task):
    job_id = "123"
    image = "my-image"
    namespace = "default"
    scratch_prefix = "s3://example-bucket/redun/"

    s3_client = boto3.client("s3", region_name="us-east-1")
    s3_client.create_bucket(Bucket="example-bucket")
    k8s_client = Mock()

    redun.executors.k8s.k8s_submit.return_value = create_job_object(uid=job_id)

    # Create example workflow script to be packaged.
    File("workflow.py").write(
        f"""
@task(load_module={custom_module})
def task1(x):
    return x + 10
    """
    )

    job = Job(a_task, a_task())
    job.id = job_id
    job.eval_hash = "eval_hash"
    code_file = package_code(scratch_prefix)
    resp = submit_task(
        k8s_client,
        image,
        namespace,
        scratch_prefix,
        job,
        a_task,
        args=(10,),
        kwargs={},
        code_file=code_file,
    )
    # # We should get a k8s job id back.
    assert resp.metadata.uid == job_id

    # # Input files should be made.
    assert File("s3://example-bucket/redun/jobs/eval_hash/input").exists()
    [code_file] = list(Dir("s3://example-bucket/redun/code"))

    # We should have submitted a job to K8S.
    redun.executors.k8s.k8s_submit.assert_called_with(
        k8s_client,
        [
            "redun",
            "--check-version",
            REDUN_REQUIRED_VERSION,
            "oneshot",
            expected_load_module,
            "--code",
            code_file.path,
            "--input",
            "s3://example-bucket/redun/jobs/eval_hash/input",
            "--output",
            "s3://example-bucket/redun/jobs/eval_hash/output",
            "--error",
            "s3://example-bucket/redun/jobs/eval_hash/error",
            a_task.name,
        ],
        image="my-image",
        namespace="default",
        job_name="k8s-job-eval_hash",
        array_size=0,
        secret_name=None,
    )


@use_tempdir
@mock_s3
@mock_k8s
@patch("redun.executors.k8s.k8s_submit")
def test_submit_task_deep_file(k8s_submit_mock):
    """
    Executor should be able to submit a task defined in a deeply nested file path.
    """
    job_id = "123"
    image = "my-image"
    namespace = "default"
    scratch_prefix = "s3://example-bucket/redun/"

    client = boto3.client("s3", region_name="us-east-1")
    client.create_bucket(Bucket="example-bucket")
    k8s_client = Mock()

    redun.executors.k8s.k8s_submit.return_value = {"jobId": "k8s-job-id"}

    # Create example workflow script to be packaged.
    File("path/to/workflow.py").write(
        """
from redun import task

@task()
def task1(x):
    return x + 10
    """
    )

    module = import_script("path/to/workflow.py")

    job = Job(module.task1, module.task1())
    job.id = job_id
    job.eval_hash = "eval_hash"
    code_file = package_code(scratch_prefix)
    resp = submit_task(
        k8s_client,
        image,
        namespace,
        scratch_prefix,
        job,
        module.task1,
        args=(10,),
        kwargs={},
        code_file=code_file,
    )

    # We should get a k8s job id back.
    assert resp["jobId"] == "k8s-job-id"

    # Input files should be made.
    assert File("s3://example-bucket/redun/jobs/eval_hash/input").exists()
    [code_file] = list(Dir("s3://example-bucket/redun/code"))

    # We should have submitted a job to k8s.
    redun.executors.k8s.k8s_submit.assert_called_with(
        k8s_client,
        [
            "redun",
            "--check-version",
            REDUN_REQUIRED_VERSION,
            "oneshot",
            "workflow",
            "--import-path",
            "path/to",
            "--code",
            code_file.path,
            "--input",
            "s3://example-bucket/redun/jobs/eval_hash/input",
            "--output",
            "s3://example-bucket/redun/jobs/eval_hash/output",
            "--error",
            "s3://example-bucket/redun/jobs/eval_hash/error",
            "task1",
        ],
        image="my-image",
        namespace="default",
        job_name="k8s-job-eval_hash",
        array_size=0,
        secret_name=None,
    )


@mock_s3
@mock_k8s
@patch("redun.executors.k8s.k8s_utils.delete_job")
@patch("redun.executors.k8s.k8s_describe_jobs")
@patch("redun.executors.k8s.get_k8s_job_pods")
@patch("redun.executors.k8s.k8s_submit")
def test_executor(
    k8s_submit_mock: Mock,
    get_k8s_job_pods_mock: Mock,
    k8s_describe_jobs_mock: Mock,
    delete_job_mock: Mock,
) -> None:
    """
    Ensure that we can submit job to K8SExecutor.
    """
    k8s_job_id = "k8s-job-id"
    k8s_job2_id = "k8s-job2-id"

    # Setup K8S mocks.
    k8s_describe_jobs_mock.return_value = iter([])
    get_k8s_job_pods_mock.return_value = iter([])

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.start()

    k8s_submit_mock.return_value = create_job_object(
        name=DEFAULT_JOB_PREFIX + "-eval_hash", uid=k8s_job_id
    )
    # Submit redun job that will succeed.
    expr = task1(10)
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"
    job.args = ((10,), {})
    executor.submit(job)

    # Wait until job arrayer actually submits it.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Ensure job options were passed correctly.
    assert k8s_submit_mock.call_args
    assert k8s_submit_mock.call_args[1] == {
        "image": "my-image",
        "namespace": "default",
        "job_name": DEFAULT_JOB_PREFIX + "-eval_hash",
        "array_size": 0,
        "retries": 1,
        "service_account_name": "default",
        "secret_name": None,
        "vcpus": 1,
        "memory": 4,
        "k8s_labels": {
            "redun_execution_id": "",
            "redun_job_id": job.id,
            "redun_project": "",
            "redun_task_name": "task1",
        },
    }

    # Setup new mock return values.
    k8s_submit_mock.return_value = create_job_object(
        name=DEFAULT_JOB_PREFIX + "-eval_hash2", uid=k8s_job2_id
    )

    # Submit redun job that will fail.
    expr2 = task1.options(memory=8)(11)
    job2 = Job(task1, expr2)
    job2.eval_hash = "eval_hash2"
    job2.args = ((11,), {})
    executor.submit(job2)

    # Let job get stale so job arrayer actually submits it.
    wait_until(lambda: executor.arrayer.num_pending == 0)

    # Ensure job options were passed correctly.
    assert k8s_submit_mock.call_args[1] == {
        "image": "my-image",
        "namespace": "default",
        "job_name": DEFAULT_JOB_PREFIX + "-eval_hash2",
        "array_size": 0,
        "retries": 1,
        "service_account_name": "default",
        "secret_name": None,
        "vcpus": 1,
        "memory": 8,
        "k8s_labels": {
            "redun_execution_id": "",
            "redun_job_id": job2.id,
            "redun_project": "",
            "redun_task_name": "task1",
        },
    }

    # Simulate k8s completing job.
    output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Simulate k8s failing.
    error = ValueError("Boom")
    error_file = File("s3://example-bucket/redun/jobs/eval_hash2/error")
    error_file.write(pickle_dumps((error, Traceback.from_error(error))), mode="wb")

    fake_k8s_job = create_job_object(uid=k8s_job_id, name=DEFAULT_JOB_PREFIX + "-eval_hash")
    fake_k8s_job.status = client.V1JobStatus(succeeded=1)
    fake_k8s_job2 = create_job_object(uid=k8s_job2_id, name=DEFAULT_JOB_PREFIX + "-eval_hash2")
    fake_k8s_job2.status = client.V1JobStatus(failed=1)
    k8s_describe_jobs_mock.return_value = [fake_k8s_job, fake_k8s_job2]

    pod = client.V1Pod(
        metadata=client.V1ObjectMeta(
            name="name",
            namespace="default",
        ),
        status=client.V1ContainerStatus(
            image=client.V1ContainerImage(),
            image_id="image_id",
            name="name",
            ready=True,
            restart_count=1,
        ),
    )
    get_k8s_job_pods_mock.return_value = [pod]

    scheduler.batch_wait([job.id, job2.id])
    executor.stop()

    # Job results and errors should be sent back to scheduler.
    assert scheduler.job_results[job.id] == 20
    assert isinstance(scheduler.job_errors[job2.id], ValueError)

    # Assert job tags.
    job.job_tags == [("k8s_job", "k8s-job-id"), ("aws_log_stream", "log1")]
    job.job_tags == [("k8s_job", "k8s-job2-id"), ("aws_log_stream", "log2")]


@mock_s3
@mock_k8s
def test_executor_handles_unrelated_jobs() -> None:
    """
    Regression test for https://insitro.atlassian.net/browse/DE-2632

    There is an expanding pattern of using a "headnode" running in batch to trigger redun
    pipelines. If the headnode and redun jobs that it spawns have a shared job_name_prefix then
    the headnode job can get gathered in the `get_jobs` call and we will try to extract the hash.
    However, since the headnode job is not a redun job, it will not have a hash and previously
    caused execution failures.

    This test confirms that jobs without hashes in their names are ignored which allows headnode
    jobs(triggered via lambda or otherwise) to share job name prefixes with the redun jobs that
    they spawn.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)

    prefix = "liveratlas_spearmancor"
    hash1 = "123456789"
    hash2 = "987654321"

    # Set up mocks to include a headnode job(no hash) and some redun jobs that it "spawned".
    executor.get_jobs.return_value = [  # type: ignore
        create_job_object(uid="headnode", name=f"{prefix}_automation_headnode"),
        create_job_object(uid="preprocess", name=f"{prefix}_preprocess-{hash1}"),
        create_job_object(uid="decode", name=f"{prefix}_decode-{hash2}"),
    ]

    executor.gather_inflight_jobs()

    assert executor.preexisting_k8s_jobs == {
        hash1: "preprocess",
        hash2: "decode",
    }


@mock_s3
@mock_k8s
@patch("redun.executors.k8s.package_code")
def test_code_packaging(package_code_mock: Mock) -> None:
    """
    Ensure that code packaging only happens on first submission.
    """
    package_code_mock.return_value = "s3://fake-bucket/code.tar.gz"

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler, code_package=True)
    executor.start()

    # Starting the executor should not have triggered code packaging.
    assert executor.code_file is None
    assert package_code_mock.call_count == 0

    # Hand create jobs.
    job1 = Job(task1, task1(10))
    job1.id = "1"
    job1.eval_hash = "eval_hash"
    job1.args = ((10,), {})

    job2 = Job(task1, task1(20))
    job2.id = "2"
    job2.eval_hash = "eval_hash"
    job2.args = ((20,), {})

    # Submit a job and ensure that the code was packaged.
    executor.submit(job1)
    assert executor.code_file
    assert package_code_mock.call_count == 1

    # Submit another job and ensure that code was not packaged again.
    executor.submit(job2)
    assert package_code_mock.call_count == 1

    executor.stop()


@mock_s3
@mock_k8s
@patch("redun.executors.k8s.k8s_describe_jobs")
@pytest.mark.skip(reason="not working")
def test_inflight_join_only_on_first_submission(k8s_describe_jobs_mock: Mock) -> None:
    """
    Ensure that inflight jobs are only gathered once and not on every job submission.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)

    executor.start()

    # Hand create jobs.
    job1 = Job(task1, task1(10))
    job1.id = "1"
    job1.eval_hash = "eval_hash"

    job2 = Job(task1, task1(20))
    job2.id = "2"
    job2.eval_hash = "eval_hash"

    # Submit redun job.
    executor.submit(job1)

    # # Ensure that inflight jobs were gathered.
    assert executor.get_jobs.call_count == 1  # type: ignore

    # # Submit the second job and confirm that job reuniting was not done again.
    executor.submit(job2)
    assert executor.get_jobs.call_count == 1  # type: ignore

    executor.stop()


@mock_s3
@mock_k8s
@patch("redun.executors.k8s.k8s_describe_jobs")
@patch("redun.executors.k8s_utils.delete_job")
@patch("redun.executors.k8s.k8s_submit")
@pytest.mark.skip(reason="not working")
def test_executor_inflight_job(
    k8s_submit_mock: Mock,
    delete_job_mock: Mock,
    k8s_describe_jobs_mock: Mock,
) -> None:
    """
    Ensure we reunite with an inflight job.
    """
    k8s_job_id = "333"

    # Setup k8s mocks.

    k8s_describe_jobs_mock.return_value = iter([])
    k8s_job = create_job_object(uid=k8s_job_id, name=DEFAULT_JOB_PREFIX + "-eval_hash")
    k8s_describe_jobs_mock.return_value = [k8s_job]

    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.get_jobs.return_value = [k8s_job]  # type: ignore
    executor.start()

    # Hand create job.
    job = Job(task1, task1(10))
    job.id = "123"
    job.eval_hash = "eval_hash"

    # Submit redun job.
    executor.submit(job)

    # Ensure no k8s jobs were submitted.
    assert k8s_submit_mock.call_count == 0

    # Simulate K8S completing with valid value.
    output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    k8s_job.status = client.V1JobStatus(succeeded=1)
    k8s_describe_jobs_mock.return_value = [k8s_job]
    # k8s_describe_jobs_mock.return_value = [k8s_job]
    scheduler.batch_wait([job.id])
    # Simulate pre-existing job output.
    output_file = File("s3://example-bucket/redun/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Ensure redun job is completed.
    assert scheduler.job_results[job.id] == 20

    executor.stop()


@task(limits={"cpu": 1}, random_option=5)
def array_task(x):
    return x + 10


@task()
def other_task(x, y):
    return x - y


@mock_s3
@mock_k8s
@patch("redun.executors.k8s_utils.delete_job")
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.k8s.k8s_describe_jobs")
@patch("redun.executors.k8s.submit_task")
@pytest.mark.skip(reason="not working")
def test_jobs_are_arrayed(
    submit_task_mock: Mock,
    k8s_describe_jobs_mock: Mock,
    get_aws_user_mock: Mock,
    delete_job_mock: Mock,
) -> None:
    """
    Tests repeated jobs are submitted as a single array job. Checks that
    job ID for the array job and child jobs end up tracked
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.arrayer.min_array_size = 3
    executor.arrayer.max_array_size = 7

    k8s_describe_jobs_mock.return_value = iter([])
    faj = create_job_object(uid="first-array-job")
    faj.spec.parallelism = 3
    faj.spec.completions = 3
    faj.spec.completion_mode = "Indexed"

    saj = create_job_object(uid="second-array-job")
    saj.spec.parallelism = 7
    saj.spec.completions = 7
    saj.spec.completion_mode = "Indexed"

    redun.executors.k8s.submit_task.side_effect = [  # type: ignore
        faj,
        saj,
        create_job_object(uid="single-job"),
    ]

    test_jobs = []
    for i in range(10):
        job = Job(array_task, array_task(i))
        job.id = f"task_{i}"
        job.eval_hash = f"eval_hash_{i}"
        executor.submit(job)
        test_jobs.append(job)

    # Wait for jobs to get submitted from arrayer to executor.
    wait_until(lambda: len(executor.pending_k8s_jobs) == 10)

    # Two array jobs, of size 7 and 3, should have been submitted.
    pending_correct = {
        f"first-array-job:{i}": test_jobs[i] for i in range(executor.arrayer.max_array_size)
    }
    pending_correct.update(
        {
            f"second-array-job:{i}": j
            for i, j in enumerate(test_jobs[executor.arrayer.max_array_size :])
        }
    )
    assert executor.pending_k8s_jobs == pending_correct

    # Two array jobs should have been submitted
    assert submit_task_mock.call_count == 2

    # Submit a different kind of job now.
    j = Job(other_task, other_task(3, 5))
    j.id = "other_task"
    j.eval_hash = "hashbrowns"
    executor.submit(j)

    assert len(executor.arrayer.pending) == 1
    pending_correct["single-job"] = j
    wait_until(lambda: executor.pending_k8s_jobs == pending_correct)

    # Make monitor thread exit correctly
    executor.stop()


@use_tempdir
@mock_s3
@mock_k8s
@patch("redun.executors.aws_utils.get_aws_user", return_value="alice")
@patch("redun.executors.k8s.K8SExecutor._submit_single_job")
def test_array_disabling(submit_single_mock: Mock, get_aws_user_mock: Mock) -> None:
    """
    Tests setting `min_array_size=0` disables job arraying.
    """
    # Setup executor.
    config = Config(
        {
            "k8s": {
                "image": "image",
                "scratch": "scratch_prefix",
                "code_includes": "*.txt",
                "min_array_size": 0,
            }
        }
    )
    scheduler = mock_scheduler()

    executor = K8SExecutor("k8s", scheduler, config["k8s"])
    executor.get_jobs = Mock()  # type: ignore
    executor.get_jobs.return_value = []

    # Submit one test job.
    job = Job(other_task, other_task(5, 3))
    job.id = "carrots"
    job.eval_hash = "why do i always say carrots in test cases idk"
    job.args = ((5, 3), {})
    executor.submit(job)

    # Job should be submitted immediately.
    assert submit_single_mock.call_args
    assert submit_single_mock.call_args[0] == (job,)

    # Monitor thread should not run.
    assert not executor.arrayer._monitor_thread.is_alive()
    executor.stop()


@mock_s3
@mock_k8s
@use_tempdir
@patch("redun.executors.k8s.k8s_submit")
def test_array_job_scratch_setup(k8s_submit_mock: Mock) -> None:
    """
    Tests that args, kwargs, and output file paths end up
    in the correct scratch locations as the right data structure.
    """
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.scratch_prefix = "./evil\ndirectory"

    k8s_job_id = "k8s-job-id"
    jo = create_job_object(name=DEFAULT_JOB_PREFIX + "-eval_hash", uid=k8s_job_id)
    jo.spec.parallelism = 3
    jo.spec.completions = 3
    jo.spec.completion_mode = "Indexed"
    k8s_submit_mock.return_value = jo

    test_jobs = []
    for i in range(10):
        job = Job(other_task, other_task(i, y=2 * i))
        job.id = f"task_{i}"
        job.eval_hash = f"hash_{i}"
        job.args = ((i,), {"y": 2 * i})
        test_jobs.append(job)

    array_uuid = executor._submit_array_job(test_jobs)

    # Check input file is on S3 and contains list of (args, kwargs) tuples
    input_file = File(get_array_scratch_file(executor.scratch_prefix, array_uuid, SCRATCH_INPUT))
    assert input_file.exists()

    with input_file.open("rb") as infile:
        arglist, kwarglist = pickle.load(infile)
    assert arglist == [(i,) for i in range(10)]
    assert kwarglist == [{"y": 2 * i} for i in range(10)]

    # Check output paths file is on S3 and contains correct output paths
    output_file = File(get_array_scratch_file(executor.scratch_prefix, array_uuid, SCRATCH_OUTPUT))
    assert output_file.exists()
    ofiles = json.loads(output_file.read())

    assert ofiles == [
        get_job_scratch_file(executor.scratch_prefix, j, SCRATCH_OUTPUT) for j in test_jobs
    ]

    # Error paths are the same as output, basically
    error_file = File(get_array_scratch_file(executor.scratch_prefix, array_uuid, SCRATCH_ERROR))
    assert error_file.exists()
    efiles = json.loads(error_file.read())

    assert efiles == [
        get_job_scratch_file(executor.scratch_prefix, j, SCRATCH_ERROR) for j in test_jobs
    ]

    # Child job eval hashes should be present as well.
    eval_file = File(get_array_scratch_file(executor.scratch_prefix, array_uuid, SCRATCH_HASHES))
    with eval_file.open("r") as evfile:
        hashes = evfile.read().splitlines()

    assert hashes == [job.eval_hash for job in test_jobs]

    # Make monitor thread exit correctly
    executor.stop()


@mock_s3
@mock_k8s
@use_tempdir
@patch("redun.executors.k8s.k8s_submit")
def test_array_oneshot(k8s_submit_mock: Mock) -> None:
    """
    Checks array child jobs can fetch their args and kwargs, and
    put their (correct) output in the right place.
    """
    # Create a code file
    file = File("workflow.py")
    file.write(
        """
from redun import task

@task()
def other_task(x, y):
   return x - y
        """
    )
    create_tar("code.tar.gz", ["workflow.py"])
    file.remove()

    # Submit 10 jobs that will be arrayed
    scheduler = mock_scheduler()
    executor = mock_executor(scheduler)
    executor.scratch_prefix = "."

    jo = create_job_object(uid="array-job-id")
    jo.spec.parallelism = 10
    jo.spec.completions = 10
    jo.spec.completion_mode = "Indexed"
    k8s_submit_mock.return_value = jo

    test_jobs = []
    for i in range(3):
        job = Job(other_task, other_task(i, y=2 * i))
        job.id = f"task_{i}"
        job.eval_hash = f"hash_{i}"
        job.args = ((i,), {"y": 2 * i})
        test_jobs.append(job)

    array_uuid = executor._submit_array_job(test_jobs)

    # Now run 2 of those jobs and make sure they work ok
    client = RedunClient()
    array_dir = os.path.join(executor.scratch_prefix, "array_jobs", array_uuid)
    input_path = os.path.join(array_dir, SCRATCH_INPUT)
    output_path = os.path.join(array_dir, SCRATCH_OUTPUT)
    error_path = os.path.join(array_dir, SCRATCH_ERROR)
    executor.stop()

    for i in range(3):
        os.environ[job_array.AWS_ARRAY_VAR] = str(i)
        client.execute(
            [
                "redun",
                "oneshot",
                "workflow.py",
                "--code",
                "code.tar.gz",
                "--array-job",
                "--input",
                input_path,
                "--output",
                output_path,
                "--error",
                error_path,
                "other_task",
            ]
        )

        # Check output files are there
        output_file = File(
            get_job_scratch_file(
                executor.scratch_prefix,
                test_jobs[i],
                SCRATCH_OUTPUT,
            )
        )

        assert pickle.loads(cast(bytes, output_file.read("rb"))) == i - 2 * i



"""
Helper functions for Azure integration.
"""

import os

import requests

from redun.logging import logger

try:
    from azure.ai.ml.identity import AzureMLOnBehalfOfCredential
    from azure.identity import DefaultAzureCredential
except (ImportError, ModuleNotFoundError):
    pass


def is_azure_managed_node() -> bool:
    """
    Function that checks if the current compute is backed by an Azure VM.
    Returns False for local devices or VMs hosted on other clouds (EC2).
    https://learn.microsoft.com/en-us/azure/virtual-machines/instance-metadata-service?tabs=linux
    """
    imds_url = "http://169.254.169.254/metadata/instance?api-version=2021-02-01"
    headers = {"Metadata": "true"}

    try:
        response = requests.get(imds_url, headers=headers, timeout=2)
        response.raise_for_status()
        # If the request is successful, we are likely on an Azure VM
        logger.debug("This node is an Azure-managed node.")
        return True
    except requests.RequestException:
        # If the request fails, we are likely not on an Azure VM
        logger.debug("This node is not an Azure-managed node.")
        return False


def get_az_credential():
    """
    Retrieve credential object and validate if it has permissions
    for storage operations (it doesn't validate the permissions to specific
    account, only if it can get token from storage.azure.com).

    The order of getting credentials is:
     - DefaultAzureCredential. By default, it uses AZURE_CLIENT_ID to look up managed
       identity info.
     - AzureMLOnBehalfOfCredential which is available only in AML compute
       (the presence of AZUREML_WORKSPACE_ID env variable is used to determine if we
       run in AML compute).
     - ManagedIdentityCredential if DEFAULT_IDENTITY_CLIENT_ID env variable is present.
       This is the last option in case we run a task in a compute cluster with default
       assigned identity and we want to override it.
    """

    if is_azure_managed_node():
        mi_client_id = (
            # the default name of the env variable, used by most tools
            os.environ.get("AZURE_CLIENT_ID", None)
            or
            # passed to AML job via ManagedIdentityConfiguration
            os.environ.get("AZUREML_USER_MANAGED_CLIENT_ID", None)
            or
            # available in AML clusters with user-assigned MI
            os.environ.get("DEFAULT_IDENTITY_CLIENT_ID", None)
        )

        # try AzureMLOnBehalfOfCredential - it doesn't use exception classes
        # from azure.identity package, so it's not compatible with ChainedTokenCredentials
        # and we need to check beforehand if it's able to get any tokens
        cr = AzureMLOnBehalfOfCredential()
        try:
            cr.get_token("https://storage.azure.com")
        except:  # noqa: E722
            cr = DefaultAzureCredential(managed_identity_client_id=mi_client_id)
    else:
        # don't attempt MI or AML credentials outside Azure
        cr = DefaultAzureCredential(exclude_managed_identity_credential=True)

    # validate if the credential can get tokens and fail early if not
    cr.get_token("https://storage.azure.com")

    return cr



import os
import re
from typing import Optional
from urllib.parse import urlparse, urlunparse

from redun.config import Config, create_config_section
from redun.file import Dir

REDUN_CONFIG_ENV = "REDUN_CONFIG"
REDUN_USER_ENV = "REDUN_USER"
REDUN_CONFIG_DIR = ".redun"
REDUN_INI_FILE = "redun.ini"
DEFAULT_REPO_NAME = "default"
DEFAULT_DB_URI = "sqlite:///redun.db"
DEFAULT_POSTGRESQL_PORT = 5432
DEFAULT_REDUN_INI = """\
# redun configuration.

[backend]
db_uri = {db_uri}

[executors.default]
type = local
max_workers = 20

# [executors.batch]
# type = aws_batch
#
# # Required:
# image =
# queue =
# s3_scratch =
#
# # Optional:
# aws_region =
# role =
# debug = False
"""


def get_abs_url(uri: str, base: str) -> str:
    """
    Returns URI with absolute path.
    """
    if re.match(r"^[^:]+:////", uri):
        # URI starts with four slashes, so it is already absolute.
        return uri

    # Parse db_uri.
    url_parts = urlparse(uri)
    if url_parts.netloc != "":
        # Non-file based URL. Use as is.
        return uri

    # Construct absolute path.
    path = url_parts.path[1:]
    abs_path = os.path.join(base, path)

    # Format a new URL with absolute path.
    abs_url_parts = url_parts._replace(path=abs_path, netloc="")
    abs_uri_prep = urlunparse(abs_url_parts)

    # Fix leading slashes to match sqlalchemy scheme (4 slashes for abs path).
    abs_uri = re.sub(r"^([^:]+):", r"\1:///", abs_uri_prep)
    return abs_uri


def get_abs_config_dir(config_dir: str, cwd: Optional[str] = None) -> str:
    """
    Returns absolute path to config_dir.

    If `config_dir` is a relative path, it is assumed to be relative to `cwd`.
    `cwd` defaults to the current working directory.
    """
    if not cwd:
        cwd = os.getcwd()
    return os.path.normpath(os.path.join(cwd, config_dir))


def get_abs_db_uri(db_uri: str, config_dir: str, cwd: Optional[str] = None) -> str:
    """
    Returns DB_URI with absolute path.

    If `db_uri` is a relative path, it is assumed to be relative to `config_dir`.
    `config_dir` itself may be relative to the current working directory (cwd).
    """
    abs_config_dir = get_abs_config_dir(config_dir, cwd)
    abs_db_uri = get_abs_url(db_uri, abs_config_dir)
    return abs_db_uri


def postprocess_config(config: Config, config_dir: str) -> Config:
    """
    Postprocess config.

    Parameters
    ----------
    config : Config
        The body of the config to update. Mutated by this operation.
    config_dir : str
        The location the config came from, for the purposes of recording into the config.
    """
    # Add default repository if not specified in config.
    default_repo = create_config_section({"config_dir": config_dir})
    if config.get("repos"):
        config["repos"][DEFAULT_REPO_NAME] = default_repo
    else:
        config["repos"] = {DEFAULT_REPO_NAME: default_repo}

    # Set default db_uri if not specified in config.
    if not config.get("backend"):
        config["backend"] = create_config_section()
    if not config["backend"].get("db_uri"):
        config["backend"]["db_uri"] = DEFAULT_DB_URI

    # Grandfather old default db_uri.
    if config["backend"]["db_uri"] == "sqlite:///.redun/redun.db":
        config["backend"]["db_uri"] = DEFAULT_DB_URI

    # Convert db_uri to absolute path.
    config["backend"]["db_uri"] = get_abs_db_uri(config["backend"]["db_uri"], config_dir)

    # Populate config_dir in backend section.
    config["backend"]["config_dir"] = config_dir

    # Create scheduler section if needed
    if not config.get("scheduler"):
        config["scheduler"] = create_config_section()

    # Convert context_file to absolute path.
    context_file = config["scheduler"].get("context_file")
    if context_file and not os.path.isabs(context_file):
        config["scheduler"]["context_file"] = os.path.join(
            get_abs_config_dir(config_dir), context_file
        )

    # Merge executors and federated tasks from any federated imported configs. Throw errors
    # on naming conflicts.
    if not config.get("executors"):
        config["executors"] = {}
    if not config.get("federated_tasks"):
        config["federated_tasks"] = {}

    executors_config = config.get("executors")
    federated_tasks_configs = config.get("federated_tasks")

    federated_config_dirs = config["scheduler"].get("federated_configs", "").strip().split("\n")

    # Skip over any empty strings
    for federated_config_dir in (x for x in federated_config_dirs if x):
        federated_config_file = Dir(federated_config_dir).file(REDUN_INI_FILE)
        federated_config = Config()
        federated_config.read_path(federated_config_file.path)

        federated_executors = federated_config.get("executors", {})
        for executor_name, executor_config in federated_executors.items():
            assert (
                executors_config.get(executor_name) is None
            ), f"Imported executor name `{executor_name}` is a duplicate."
            executors_config[executor_name] = executor_config

        federated_entrypoints = federated_config.get("federated_tasks", {})
        for entrypoint_name, entrypoint_config in federated_entrypoints.items():
            assert (
                federated_tasks_configs.get(entrypoint_name) is None
            ), f"Imported federated_task name `{entrypoint_name}` is a duplicate."
            federated_tasks_configs[entrypoint_name] = entrypoint_config

    return config



from typing import Dict, Iterator, Optional, Tuple

from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Footer, Static

from redun.backends.db import CallNode, Value
from redun.backends.db.dataflow import (
    ArgumentValue,
    CallNodeValue,
    DataflowDOM,
    DataflowNode,
    DataflowSectionDOM,
    get_default_arg_name,
    get_node_hash,
    get_task_args,
    make_dataflow_dom,
    walk_dataflow,
)
from redun.console.screens import RedunScreen
from redun.utils import assert_never, trim_string


def display_node(node: Optional[DataflowNode], renames: Dict[str, str]) -> Tuple[str, str]:
    """
    Formats a dataflow node to a string.
    """
    if isinstance(node, CallNode):
        return ("", display_call_node(node, renames))

    elif isinstance(node, ArgumentValue):
        return (
            "argument of",
            display_call_node(node.argument.call_node, renames),
        )

    elif isinstance(node, CallNodeValue):
        return ("", display_call_node(node.call_node, renames))

    elif node is None:
        return ("", "origin")

    else:
        assert_never(node)


def display_call_node(call_node: CallNode, renames: Dict[str, str]) -> str:
    """
    Formats a CallNode to a string.
    """
    try:
        arg_names = get_task_args(call_node.task)
    except SyntaxError:
        arg_names = [get_default_arg_name(i) for i in range(len(call_node.arguments))]
    args = [renames.get(arg, arg) for arg in arg_names]
    return "{task_name}({args})".format(task_name=call_node.task.name, args=", ".join(args))


def display_value(value: Value) -> str:
    """
    Format a Value to a string.
    """
    return trim_string(repr(value.preview).replace("\n", " "))


def display_hash(node: Optional[DataflowNode]) -> str:
    """
    Formats hash for a DataflowNode.
    """
    node_hash = get_node_hash(node)
    if node_hash:
        return f"<{node_hash[:8]}> "
    else:
        return ""


def display_link(hash: Optional[str], text: str) -> str:
    if hash:
        return f"[@click=screen.click_hash('{hash}')]{text}[/]"
    else:
        return text


def display_section(dom: DataflowSectionDOM) -> Iterator[str]:
    """
    Yields lines for displaying a dataflow section DOM.
    """

    # Display assign line.
    assign = dom.assign
    yield "{var_name} [bold]<--[/] {prefix}{link}".format(
        var_name=assign.var_name,
        prefix=assign.prefix + " " if assign.prefix else "",
        link=display_link(
            get_node_hash(assign.node), display_hash(assign.node) + assign.node_display
        ),
    )

    indent = len(assign.var_name) + 1

    # Display routing.
    for routing_def in dom.routing:
        yield "{indent}[bold]<--[/] {prefix}{link}".format(
            indent=" " * indent,
            prefix=routing_def.prefix + " " if routing_def.prefix else "",
            link=display_link(
                get_node_hash(routing_def.node),
                display_hash(routing_def.node) + routing_def.node_display,
            ),
        )

    # Display argument definitions.
    if dom.args:
        max_var_len = max(len(arg.var_name) for arg in dom.args)
        for arg in dom.args:
            yield "  {var_name}{padding} [bold]=[/] {link}".format(
                var_name=arg.var_name,
                padding=" " * (max_var_len - len(arg.var_name)),
                link=display_link(
                    arg.value.value_hash,
                    display_hash(arg.value) + display_value(arg.value),
                ),
            )


def display_dataflow(dom: DataflowDOM) -> Iterator[str]:
    """
    Yields for lines displaying a dataflow DOM.
    """
    for dom_section in dom:
        yield from display_section(dom_section)
        yield ""


class UpstreamDataflowScreen(RedunScreen):
    """
    Screen for exploring the upstream dataflow of a redun Value.
    """

    path_pattern = "^upstreams/(.*)$"

    BINDINGS = [
        ("q", "quit", "Quit"),
        ("escape", "back", "Back"),
        ("r", "repl", "REPL"),
    ]

    def __init__(self, value_hash: str):
        self.value_hash = value_hash
        super().__init__()

        self.value = self.app.session.get(Value, value_hash)

    def get_path(self) -> str:
        return f"upstreams/{self.value_hash}"

    def action_click_value(self, value_hash: str) -> None:
        self.app.route([f"values/{value_hash}"])

    def action_click_hash(self, hash: str) -> None:
        self.app.route([hash])

    def action_repl(self) -> None:
        screen = self.app.get_screen("ReplScreen")
        screen.update({"value": self.value.value_parsed}, obj_id=self.value_hash)
        self.app.push_screen(screen)

    def compose(self) -> ComposeResult:
        if self.value is None:
            yield Container(
                self.header,
                Static(f"[bold red]Unknown value {self.value_hash}[/]"),
                Footer(),
            )
            return

        edges = walk_dataflow(self.app.scheduler.backend, self.value)
        dom = make_dataflow_dom(edges, new_varname="value")
        lines = display_dataflow(dom)

        yield Container(
            self.header,
            Container(
                Static(f"[bold]Upstream dataflow[/] {self.value_hash}"),
                Static(),
                Static(
                    f"[bold]value [b]=[/] [@click=screen.click_value('{self.value_hash}')]{display_hash(self.value)}{repr(self.value.preview)}[/]",  # noqa: E501
                    classes="dataflow",
                ),
                Static(),
                Static("\n".join(lines), classes="dataflow"),
                id="upstreams-screen",
            ),
            Footer(),
        )



import os
from typing import Any, Dict, List, Optional, Tuple

from redun.executors.scratch import (
    SCRATCH_ERROR,
    SCRATCH_INPUT,
    SCRATCH_OUTPUT,
    SCRATCH_STATUS,
    get_array_scratch_file,
    get_job_scratch_file,
)
from redun.file import File, get_filesystem
from redun.scheduler import Job
from redun.task import CacheScope, Task
from redun.utils import get_import_paths, pickle_dump

REDUN_PROG = "redun"
REDUN_REQUIRED_VERSION = ">=0.4.1"


def get_oneshot_command(
    scratch_prefix: str,
    job: Job,
    a_task: Task,
    args: Tuple = (),
    kwargs: Dict[str, Any] = {},
    job_options: dict = {},
    code_file: Optional[File] = None,
    array_uuid: Optional[str] = None,
    input_path: Optional[str] = None,
    output_path: Optional[str] = None,
    error_path: Optional[str] = None,
) -> List[str]:
    """
    Returns a redun oneshot command for a Job.
    """
    if array_uuid:
        if not input_path:
            input_path = get_array_scratch_file(scratch_prefix, array_uuid, SCRATCH_INPUT)
        if not output_path:
            output_path = get_array_scratch_file(scratch_prefix, array_uuid, SCRATCH_OUTPUT)
        if not error_path:
            error_path = get_array_scratch_file(scratch_prefix, array_uuid, SCRATCH_ERROR)

        # Assume arguments are already serialized.
    else:
        if not input_path:
            input_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_INPUT)
        if not output_path:
            output_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_OUTPUT)
        if not error_path:
            error_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_ERROR)

        # Serialize arguments to input file.
        # Array jobs set this up themselves.
        input_file = File(input_path)
        with input_file.open("wb") as out:
            pickle_dump([args, kwargs], out)

    # Determine additional python import paths.
    import_args = []
    base_path = os.getcwd()
    for abs_path in get_import_paths():
        # Use relative paths so that they work inside the docker container.
        rel_path = os.path.relpath(abs_path, base_path)
        import_args.append("--import-path")
        import_args.append(rel_path)

    # Build job command.
    code_arg = ["--code", code_file.path] if code_file else []
    array_arg = ["--array-job"] if array_uuid else []
    cache_arg = (
        []
        if CacheScope(job_options.get("cache_scope", CacheScope.BACKEND)) == CacheScope.BACKEND
        else ["--no-cache"]
    )
    command = (
        [
            REDUN_PROG,
            "--check-version",
            REDUN_REQUIRED_VERSION,
            "oneshot",
            a_task.load_module,
        ]
        + import_args
        + code_arg
        + array_arg
        + cache_arg
        + [
            "--input",
            input_path,
            "--output",
            output_path,
            "--error",
            error_path,
            a_task.fullname,
        ]
    )
    return command


def get_script_task_command(
    scratch_prefix: str,
    job: Job,
    command: str,
    exit_command: str = "",
    as_mount: bool = False,
) -> List[str]:
    """
    Returns a shell script to run a script task.
    """
    input_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_INPUT)
    output_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_OUTPUT)
    error_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_ERROR)
    status_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_STATUS)

    # Serialize arguments to input file.
    File(input_path).write(command)

    input_stage = File(input_path).stage(".task_command").render_stage(as_mount)
    output_unstage = File(output_path).stage(".task_output").render_unstage(as_mount)
    error_unstage = File(error_path).stage(".task_error").render_unstage(as_mount)
    status_unstage = get_filesystem(url=status_path).shell_copy(
        None, status_path, as_mount=as_mount
    )

    return [
        "bash",
        "-c",
        "-o",
        "pipefail",
        f"""
{input_stage}
chmod +x .task_command
(
  ./.task_command \
  2> >(tee .task_error >&2) | tee .task_output
) && (
    {output_unstage}
    {error_unstage}
    echo ok | {status_unstage}
) || (
    [ -f .task_output ] && {output_unstage}
    [ -f .task_error ] && {error_unstage}
    echo fail | {status_unstage}
    {exit_command}
)
""",
    ]



import json
import logging
import re
import threading
import time
import uuid
from collections import OrderedDict
from configparser import SectionProxy
from typing import Any, Dict, List, Optional, cast

from google.api_core.exceptions import NotFound
from google.cloud.batch_v1 import JobStatus, Task, TaskStatus

from redun.executors import gcp_utils
from redun.executors.base import Executor, register_executor
from redun.executors.code_packaging import package_code, parse_code_package_config
from redun.executors.command import get_oneshot_command, get_script_task_command
from redun.executors.docker import DockerExecutor, get_docker_executor_config
from redun.executors.scratch import (
    SCRATCH_ERROR,
    SCRATCH_HASHES,
    SCRATCH_INPUT,
    SCRATCH_OUTPUT,
    get_array_scratch_file,
    get_job_scratch_dir,
    get_job_scratch_file,
    parse_job_error,
    parse_job_result,
)
from redun.file import File
from redun.job_array import JobArrayer
from redun.scheduler import Job as RedunJob
from redun.scheduler import Scheduler
from redun.scripting import DEFAULT_SHELL, get_task_command
from redun.task import CacheScope
from redun.utils import pickle_dump

REDUN_ARRAY_JOB_PREFIX = "redun-array-"
REDUN_JOB_PREFIX = "redun-"

REDUN_HASH_LABEL_KEY = "redun_hash"
REDUN_JOB_TYPE_LABEL_KEY = "redun_job_type"

DEBUG_SCRATCH = "scratch"


@register_executor("gcp_batch")
class GCPBatchExecutor(Executor):
    """
    A redun Executor for running jobs on GCP Batch.
    """

    def __init__(
        self,
        name: str,
        scheduler: Optional["Scheduler"] = None,
        config: Optional[SectionProxy] = None,
    ):
        super().__init__(name, scheduler=scheduler)
        if config is None:
            raise ValueError("GCPBatchExecutor requires config.")

        self.debug = config.getboolean("debug", fallback=False)

        # Use DockerExecutor for local debug mode.
        docker_config = get_docker_executor_config(config)
        docker_config["scratch"] = config.get("debug_scratch", fallback=DEBUG_SCRATCH)
        docker_config["include_aws_env"] = "False"
        self._docker_executor = DockerExecutor(name + "_debug", scheduler, config=docker_config)

        self.gcp_batch_client = gcp_utils.get_gcp_batch_client()
        self.gcp_compute_client = gcp_utils.get_gcp_compute_client()

        # Required config.
        self.project = config["project"]
        self.region = config["region"]
        self.gcs_scratch_prefix = config["gcs_scratch"]
        self.image = config["image"]

        # Optional config
        self.code_package = parse_code_package_config(config)
        self.code_file: Optional[File] = None

        # Monitoring internals
        self.preexisting_batch_tasks: Dict[str, str] = {}  # RedunJob hash -> BatchJob task name
        # We use an OrderedDict in order to retain submission order.
        self.pending_batch_tasks: Dict[str, "RedunJob"] = OrderedDict()
        self.is_running = False
        self.interval = config.getfloat("job_monitor_interval", fallback=5.0)
        self._thread: Optional[threading.Thread] = None
        self.arrayer = JobArrayer(
            self._submit_jobs,
            self._on_error,
            submit_interval=self.interval,
            stale_time=config.getfloat("job_stale_time", fallback=3.0),
            min_array_size=config.getint("min_array_size", fallback=2),
            max_array_size=config.getint("max_array_size", fallback=5000),
        )

        # Default task options.
        self.default_task_options: Dict[str, Any] = {
            "machine_type": config.get("machine_type", fallback="e2-standard-4"),
            "task_count": config.getint("task_count", fallback=1),
            "retries": config.getint("retries", fallback=2),
            "priority": config.getint("priority", fallback=30),
            "service_account_email": config.get("service_account_email", fallback=""),
        }
        if config.get("labels"):
            self.default_task_options["labels"] = json.loads(config.get("labels"))

    def _on_error(self, error: Exception) -> None:
        """
        Callback for JobArrayer to report scheduler-level errors.
        """
        assert self._scheduler
        self._scheduler.reject_job(None, error)

    def set_scheduler(self, scheduler: Scheduler) -> None:
        super().set_scheduler(scheduler)
        self._docker_executor.set_scheduler(scheduler)

    def _is_debug_job(self, job: RedunJob) -> bool:
        """
        Returns True if job should be sent to debugging DockerExecutor.
        """
        return self.debug or job.get_option("debug", False)

    def gather_inflight_jobs(self) -> None:
        batch_jobs = gcp_utils.list_jobs(
            client=self.gcp_batch_client, project_id=self.project, region=self.region
        )
        inflight_job_statuses = [
            JobStatus.State.SCHEDULED,
            JobStatus.State.QUEUED,
            JobStatus.State.RUNNING,
        ]
        inflight_task_statuses = [
            TaskStatus.State.ASSIGNED,
            TaskStatus.State.PENDING,
            TaskStatus.State.RUNNING,
        ]

        for job in batch_jobs:
            # If a job is not labelled with a redun hash then it is not a redun job
            if REDUN_HASH_LABEL_KEY not in job.labels:
                continue

            if job.status.state not in inflight_job_statuses:
                continue

            job_hash = job.labels[REDUN_HASH_LABEL_KEY]

            # Single jobs have only one task in their task group and can be simply added to dict of
            # pre-existing jobs.
            if job.task_groups[0].task_count == 1:
                task_name = f"{job.task_groups[0].name}/tasks/0"
                self.preexisting_batch_tasks[job_hash] = task_name

            # Otherwise we need to read the eval file to reunite with batch tasks in the batch job
            else:
                eval_file = File(
                    get_array_scratch_file(self.gcs_scratch_prefix, job_hash, SCRATCH_HASHES)
                )
                if not eval_file.exists():
                    # Eval file does not exist, so we cannot reunite with this array job.
                    continue

                # Get eval_hash for all jobs that were part of the array
                eval_hashes = cast(str, eval_file.read("r")).splitlines()

                batch_tasks = gcp_utils.list_tasks(
                    client=self.gcp_batch_client, group_name=job.task_groups[0].name
                )
                for array_index, task in enumerate(batch_tasks):
                    # Skip job if it is not in one of the 'inflight' states
                    if task.status.state in inflight_task_statuses:
                        array_job_hash = eval_hashes[array_index]
                        self.preexisting_batch_tasks[array_job_hash] = task.name

    def _get_job_options(self, job: RedunJob, array_uuid: Optional[str] = None) -> dict:
        """
        Determine the task options for a job.

        Task options can be specified at the job-level have precedence over
        the executor-level (within `redun.ini`):
        """
        assert job.eval_hash

        job_options = job.get_options()

        task_options = {
            **self.default_task_options,
            **job_options,
        }

        default_tags = {
            REDUN_HASH_LABEL_KEY: array_uuid if array_uuid else job.eval_hash,
            REDUN_JOB_TYPE_LABEL_KEY: "script" if job.task.script else "container",
        }

        project = self.project
        region = self.region
        machine_type = task_options.get("machine_type")

        # If either memory or vCPUS are not provided use the maximum available based on machine
        # type otherwise use the min of what is available and what is requested
        compute_machine_type = gcp_utils.get_compute_machine_type(
            self.gcp_compute_client, project, region, machine_type
        )

        requested_memory = task_options.get("memory", 0)
        # MiB to GB
        max_memory = compute_machine_type.memory_mb / 1024
        if requested_memory <= 0:
            task_options["memory"] = max_memory
        else:
            task_options["memory"] = min(requested_memory, max_memory)

        requested_vcpus = task_options.get("vcpus", 0)
        max_vcpus = compute_machine_type.guest_cpus
        if requested_vcpus <= 0:
            task_options["vcpus"] = max_vcpus
        else:
            task_options["vcpus"] = min(requested_vcpus, max_vcpus)

        # Merge labels if needed.
        labels: Dict[str, str] = {
            **self.default_task_options.get("labels", {}),
            **job_options.get("labels", {}),
            **default_tags,
        }

        task_options["labels"] = labels

        return task_options

    def submit(self, job: RedunJob) -> None:
        """
        Submit RedunJob to executor.
        """
        assert job.task
        assert not job.task.script

        if self._is_debug_job(job):
            return self._docker_executor.submit(job)
        else:
            self._submit(job)

    def submit_script(self, job: RedunJob) -> None:
        """
        Submit RedunJob for script task to executor.
        """
        assert job.task
        assert job.task.script

        if self._is_debug_job(job):
            return self._docker_executor.submit_script(job)
        else:
            self._submit(job)

    def _submit(self, job: RedunJob) -> None:
        """
        Submit RedunJob to executor.
        """

        assert self._scheduler
        assert job.task
        assert not self.debug

        # We check is_running here as a way of determining whether this is the first submission
        # or not. If we are already running, then we know we have already had jobs submitted
        # and done the inflight check so no reason to do that again here.
        if not self.is_running:
            # Precompute existing inflight jobs for job reuniting.
            self.gather_inflight_jobs()

        # Package code if necessary and we have not already done so. If code_package
        # is False, then we can skip this step. Additionally, if we have already
        # packaged and set code_file, then we do not need to repackage.
        if self.code_package is not False and self.code_file is None:
            code_package = self.code_package or {}
            assert isinstance(code_package, dict)
            self.code_file = package_code(self.gcs_scratch_prefix, code_package)

        task_options = self._get_job_options(job)
        cache_scope = CacheScope(task_options.get("cache_scope", CacheScope.BACKEND))

        # Determine if we can reunite with a previous Batch output or job.
        batch_task_name: Optional[str] = None
        if cache_scope == CacheScope.BACKEND and job.eval_hash in self.preexisting_batch_tasks:
            batch_task_name = self.preexisting_batch_tasks.pop(job.eval_hash)

            job_dir = get_job_scratch_dir(self.gcs_scratch_prefix, job)
            existing_task = gcp_utils.get_task(
                client=self.gcp_batch_client, task_name=batch_task_name
            )

            if existing_task:
                self.log(
                    "reunite redun job {redun_job} with {job_type}:\n"
                    "  task_name        = {batch_task_name}\n"
                    "  gcs_scratch_path = {job_dir}".format(
                        redun_job=job.id,
                        job_type="GCP Batch Job Task",
                        batch_task_name=batch_task_name,
                        job_dir=job_dir,
                    )
                )
                self.pending_batch_tasks[existing_task.name] = job
        if batch_task_name is None:
            self.arrayer.add_job(job)

        self._start()

    def _submit_jobs(self, jobs: List[RedunJob]) -> None:
        """
        Callback for JobArrayer to return arrays of Jobs.
        """
        if len(jobs) == 1:
            # Singleton jobs are given as single element lists.
            self._submit_single_job(jobs[0])
        else:
            self._submit_array_job(jobs)

    def _submit_array_job(self, jobs: List[RedunJob]) -> str:
        """
        Submits an array job, returning job name uuid
        """
        array_size = len(jobs)
        all_args = []
        all_kwargs = []
        for job in jobs:
            assert job.args
            all_args.append(job.args[0])
            all_kwargs.append(job.args[1])

        # All jobs identical so just grab the first one
        job = jobs[0]
        assert job.task
        if job.task.script:
            raise NotImplementedError("Array jobs not supported for scripts")

        # Generate a unique name for job with no '-' to simplify job name parsing.
        array_uuid = str(uuid.uuid4())

        task_options = self._get_job_options(job, array_uuid=array_uuid)

        # Setup input, output and error path files. Input file is a pickled list
        # of args, and kwargs, for each child job.
        input_file = get_array_scratch_file(self.gcs_scratch_prefix, array_uuid, SCRATCH_INPUT)
        with File(input_file).open("wb") as out:
            pickle_dump([all_args, all_kwargs], out)

        # Output file is a plaintext list of output paths, for each child job.
        output_file = get_array_scratch_file(self.gcs_scratch_prefix, array_uuid, SCRATCH_OUTPUT)
        output_paths = [
            get_job_scratch_file(self.gcs_scratch_prefix, job, SCRATCH_OUTPUT) for job in jobs
        ]
        with File(output_file).open("w") as ofile:
            json.dump(output_paths, ofile)

        # Error file is a plaintext list of error paths, one for each child job.
        error_file = get_array_scratch_file(self.gcs_scratch_prefix, array_uuid, SCRATCH_ERROR)
        error_paths = [
            get_job_scratch_file(self.gcs_scratch_prefix, job, SCRATCH_ERROR) for job in jobs
        ]
        with File(error_file).open("w") as efile:
            json.dump(error_paths, efile)

        # Eval hash file is plaintext hashes of child jobs for matching for job
        # reuniting.
        eval_file = get_array_scratch_file(self.gcs_scratch_prefix, array_uuid, SCRATCH_HASHES)
        with File(eval_file).open("w") as eval_f:
            eval_f.write("\n".join([job.eval_hash for job in jobs]))  # type: ignore

        image = task_options.pop("image", self.image)
        project = task_options.pop("project", self.project)
        region = task_options.pop("region", self.region)

        # Drop task_count, since we can infer it from the number of jobs.
        task_options.pop("task_count", None)
        command = get_oneshot_command(
            self.gcs_scratch_prefix,
            job,
            job.task,
            code_file=self.code_file,
            array_uuid=array_uuid,
        )

        gcp_job = gcp_utils.batch_submit(
            client=self.gcp_batch_client,
            job_name=f"{REDUN_ARRAY_JOB_PREFIX}{array_uuid}",
            project=project,
            region=region,
            image=image,
            commands=command,
            gcs_scratch_prefix=self.gcs_scratch_prefix,
            task_count=array_size,
            **task_options,
        )

        gcp_job_task_count = gcp_job.task_groups[0].task_count
        if gcp_job_task_count != array_size:
            raise AssertionError(
                f"Batch job task group has {gcp_job_task_count} tasks "
                f"instead of requested array size {array_size}."
            )

        array_task_group_name = gcp_job.task_groups[0].name
        for i in range(gcp_job_task_count):
            self.pending_batch_tasks[f"{array_task_group_name}/tasks/{i}"] = jobs[i]

        self.log(
            "submit {array_size} redun job(s) as {job_type} {array_job_id}:\n"
            "  array_job_id          = {array_job_id}\n"
            "  array_task_group_name = {array_task_group_name}\n"
            "  array_size            = {array_size}\n"
            "  gcs_scratch_path      = {job_dir}\n".format(
                array_size=array_size,
                job_type="GCP Batch Job",
                array_job_id=gcp_job.uid,
                array_task_group_name=array_task_group_name,
                job_dir=get_array_scratch_file(self.gcs_scratch_prefix, array_uuid, ""),
            )
        )

        return array_uuid

    def _submit_single_job(self, job: RedunJob) -> None:
        """
        Actually submits a job. Caching detects if it should be part
        of an array job
        """
        assert job.task
        assert job.eval_hash
        assert self.project
        assert self.region
        assert self.gcs_scratch_prefix

        assert job.args
        args, kwargs = job.args

        task_options = self._get_job_options(job)
        image = task_options.pop("image", self.image)
        project = task_options.pop("project", self.project)
        region = task_options.pop("region", self.region)

        # Submit a new Batch job.
        gcp_job = None
        if not job.task.script:
            command = get_oneshot_command(
                self.gcs_scratch_prefix,
                job,
                job.task,
                args=args,
                kwargs=kwargs,
                code_file=self.code_file,
            )

            gcp_job = gcp_utils.batch_submit(
                client=self.gcp_batch_client,
                job_name=f"{REDUN_JOB_PREFIX}{job.id}",
                project=project,
                region=region,
                image=image,
                commands=command,
                gcs_scratch_prefix=self.gcs_scratch_prefix,
                **task_options,
            )
        else:
            task_command = get_task_command(job.task, args, kwargs)
            script_command = get_script_task_command(
                self.gcs_scratch_prefix,
                job,
                task_command,
                exit_command="exit 1",
                as_mount=True,
            )

            # Get buckets - This can probably be improved.
            mount_buckets: List[str] = list(
                set(re.findall(r"/mnt/disks/([^ \/]+)/", script_command[-1] + task_command))
            )

            # Convert start command to script.
            script_path = get_job_scratch_file(self.gcs_scratch_prefix, job, ".redun_job.sh")

            File(script_path).write(DEFAULT_SHELL + "\n" + script_command[-1])

            script_path = script_path.replace("gs://", "/mnt/disks/")

            # GCP Batch takes script as a string and requires quoting of -c argument
            script_command = ["bash", script_path]
            gcp_job = gcp_utils.batch_submit(
                client=self.gcp_batch_client,
                job_name=f"redun-{job.id}",
                project=project,
                region=region,
                image=image,
                commands=script_command,
                mount_buckets=mount_buckets,
                **task_options,
            )

        job_dir = get_job_scratch_dir(self.gcs_scratch_prefix, job)
        self.log(
            "submit redun job {redun_job} as {job_type}:\n"
            "  task_name        = {task_name}\n"
            "  gcs_scratch_path = {job_dir}\n".format(
                redun_job=job.id,
                job_type="GCP Batch Job Task",
                task_name=gcp_job.task_groups[0].name,
                job_dir=job_dir,
            )
        )
        task_name = gcp_job.task_groups[0].name
        # For a single submitted task we can set task name directly
        self.pending_batch_tasks[f"{task_name}/tasks/0"] = job

    def _start(self) -> None:
        """
        Start monitoring thread.
        """
        if not self._thread or not self._thread.is_alive():
            self.is_running = True
            self._thread = threading.Thread(target=self._monitor, daemon=False)
            self._thread.start()

    def stop(self) -> None:
        """
        Stop Executor and monitoring thread.
        """
        self.is_running = False

        self._docker_executor.stop()
        self.arrayer.stop()

        # Stop monitor thread.
        if (
            self._thread
            and self._thread.is_alive()
            and threading.get_ident() != self._thread.ident
        ):
            self._thread.join()

    def _monitor(self) -> None:
        """
        Thread for monitoring running GCP Batch jobs.
        """
        assert self._scheduler

        # Need new client for thread safety
        gcp_batch_client = gcp_utils.get_gcp_batch_client()

        try:
            while self.is_running and (self.pending_batch_tasks or self.arrayer.num_pending):
                if self._scheduler.logger.level >= logging.DEBUG:
                    self.log(
                        f"Preparing {self.arrayer.num_pending} job(s) for Job Arrays.",
                        level=logging.DEBUG,
                    )
                    self.log(
                        f"Waiting on {len(self.pending_batch_tasks)} Batch tasks(s):\n\t"
                        + "\n\t".join(sorted(self.pending_batch_tasks.keys())),
                        level=logging.DEBUG,
                    )

                # Copy pending_batch_tasks.keys() since it can change due to new submissions.
                task_names = list(self.pending_batch_tasks.keys())
                for name in task_names:
                    try:
                        task = gcp_utils.get_task(client=gcp_batch_client, task_name=name)
                        self._process_task_status(task)
                    except NotFound:
                        # Batch Job has not instantiated tasks yet so ignore this NotFound error
                        continue

                time.sleep(self.interval)

        except Exception as error:
            # Since we run this is method at the top-level of a thread, we
            # need to catch all exceptions so we can properly report them to
            # the scheduler.
            self._scheduler.reject_job(None, error)

        self.log("Shutting down executor...", level=logging.DEBUG)
        self.stop()

    def _process_task_status(self, task: Task) -> None:
        assert self._scheduler
        state = task.status.state
        if state == TaskStatus.State.STATE_UNSPECIFIED:
            return
        elif state == TaskStatus.State.RUNNING:
            return
        elif state == TaskStatus.State.ASSIGNED:
            return
        elif state == TaskStatus.State.SUCCEEDED:
            redun_job = self.pending_batch_tasks.pop(task.name)
            result, exists = parse_job_result(self.gcs_scratch_prefix, redun_job)
            if exists:
                self._scheduler.done_job(redun_job, result)
            else:
                # This can happen if job ended in an inconsistent state.
                self._scheduler.reject_job(
                    redun_job,
                    FileNotFoundError(
                        get_job_scratch_file(self.gcs_scratch_prefix, redun_job, SCRATCH_OUTPUT)
                    ),
                )
        elif state == TaskStatus.State.FAILED:
            redun_job = self.pending_batch_tasks.pop(task.name)
            error, error_traceback = parse_job_error(self.gcs_scratch_prefix, redun_job)

            self._scheduler.reject_job(redun_job, error, error_traceback=error_traceback)



"""
Tests Glue helper functions in `redun.glue`.
Uses a pytest fixture to provide a spark context.
Spark imports are kept within the tests to avoid changing the behavior of NamedTuple.
"""

import pytest

from redun import glue


@pytest.fixture()
def spark_session(request):
    """
    Returns a Spark Contex
    """
    from pyspark.sql import SparkSession

    return SparkSession.builder.master("local[*]").appName("redun_test").getOrCreate()


def test_udf_without_spark():
    """
    Without a Spark context, UDF should operate like a normal Python function
    """

    from pyspark.sql.types import ArrayType, IntegerType

    @glue.udf
    def noarg_udf(input1: str):
        return input1 + " potatoes"

    @glue.udf(return_type=ArrayType(IntegerType()))
    def udf_tester_func(input1: str, input2: str):
        return [int(input1), int(input2)]

    assert udf_tester_func("3", "2") == [3, 2]
    assert noarg_udf("i like") == "i like potatoes"


def test_udf_with_spark(spark_session):
    from pyspark.sql.column import Column
    from pyspark.sql.types import ArrayType, IntegerType

    @glue.udf(return_type=ArrayType(IntegerType()))
    def udf_tester_func2(input1: str, input2: str):
        return [int(input1), int(input2)]

    @glue.udf
    def noarg_udf(input1: str):
        return input1 + "hello"

    assert isinstance(udf_tester_func2("3", "2"), Column)

    df = spark_session.createDataFrame(
        [("a", "1", "2"), ("b", "2", "3"), ("ab", "0", "0")], schema=["name", "x", "y"]
    )
    df = df.withColumn("udf_col", udf_tester_func2(df.x, df.y))
    df = df.withColumn("str_col", noarg_udf(df.name))
    assert ("udf_col", "array<int>") in df.dtypes
    assert ("str_col", "string") in df.dtypes


def test_sql_query(spark_session):
    df = spark_session.createDataFrame(
        [("a" * i + "h", i) for i in range(10)], schema=["name", "count"]
    )

    # Test invalid alias
    with pytest.raises(ValueError, match=r"alias .* not referenced"):
        _ = glue.sql_query(df, "select max(length(name)) as longest from aliased_name")

    result = glue.sql_query(
        df,
        "select max(length(name)) as longest from aliased_name",
        dataset_alias="aliased_name",
    )

    assert result.collect()[0].asDict()["longest"] == 10



"""
Utilities for working with executor scratch directories.
"""

import os
import pickle
from typing import Any, Callable, Optional, Tuple, cast

from redun.file import File
from redun.scheduler import Job, Traceback
from redun.scripting import ScriptError

# Scratch filenames.
SCRATCH_INPUT = "input"
SCRATCH_OUTPUT = "output"
SCRATCH_ERROR = "error"
SCRATCH_HASHES = "eval_hashes"
SCRATCH_STATUS = "status"


class ScratchError(Exception):
    """
    Error when reading data from scratch directory.
    """

    pass


class ExceptionNotFoundError(ScratchError):
    """
    Error when serialized exception is not found in scratch directory.
    """

    pass


def get_job_scratch_dir(scratch_prefix: str, job: Job) -> str:
    """
    Returns scratch directory for a redun Job.
    """
    assert job.eval_hash
    return os.path.join(scratch_prefix, "jobs", job.eval_hash)


def get_job_scratch_file(scratch_prefix: str, job: Job, filename: str) -> str:
    """
    Returns scratch path for a file related to a redun Job.
    """
    assert job.eval_hash
    return os.path.join(scratch_prefix, "jobs", job.eval_hash, filename)


def get_code_scratch_file(scratch_prefix: str, tar_hash: str, use_zip: bool = False) -> str:
    """
    Returns scratch path for a code package tar file.
    """
    return os.path.join(scratch_prefix, "code", tar_hash + (".zip" if use_zip else ".tar.gz"))


def get_array_scratch_file(scratch_prefix: str, job_array_id: str, filename: str) -> str:
    """
    Returns an scratch path for a file related to an array of jobs.
    """
    return os.path.join(scratch_prefix, "array_jobs", job_array_id, filename)


def get_execution_scratch_file(scratch_prefix: str, execution_id: str, filename: str) -> str:
    """
    Returns an scratch path for a sending data to and from a remote execution.
    """
    return os.path.join(scratch_prefix, "execution", execution_id, filename)


def parse_job_result(
    scratch_prefix: str,
    job: Job,
    is_valid_value: Optional[Callable[[Any], bool]] = None,
) -> Tuple[Any, bool]:
    """
    Returns job output from scratch directory.

    Returns a tuple of (result, exists).
    """
    output_file = File(get_job_scratch_file(scratch_prefix, job, SCRATCH_OUTPUT))
    if output_file.exists():
        if not job.task.script:
            with output_file.open("rb") as infile:
                result = pickle.load(infile)
        else:
            result = [0, output_file.read(mode="rb")]  # TODO: Get real exitcode.

        if is_valid_value is None or is_valid_value(result):
            return result, True
    return None, False


def parse_job_error(scratch_prefix: str, job: Job) -> Tuple[Exception, Traceback]:
    """
    Returns job error from scratch directory.
    """
    error_path = get_job_scratch_file(scratch_prefix, job, SCRATCH_ERROR)
    error_file = File(error_path)

    if not job.task.script:
        # Normal Tasks (non-script) store errors as Pickled exception, traceback tuples.
        if error_file.exists():
            try:
                error, error_traceback = pickle.loads(cast(bytes, error_file.read("rb")))
            except Exception as parse_error:
                error = ScratchError(
                    f"Error could not be parsed from scratch directory. See logs. {parse_error}"
                )
                error_traceback = Traceback.from_error(error)
        else:
            error = ExceptionNotFoundError("Exception and traceback could not be found for Job.")
            error_traceback = Traceback.from_error(error)

    else:
        # Script task.
        if error_file.exists():
            error = ScriptError(cast(bytes, error_file.read("rb")))
        else:
            error = ExceptionNotFoundError("stderr could not be found for Job.")
        error_traceback = Traceback.from_error(error)

    return error, error_traceback



from enum import Enum
from functools import lru_cache
from typing import Dict, Iterable, List, Tuple, Union, Optional

from google.api_core import gapic_v1
from google.cloud import batch_v1, compute_v1

from redun.version import version


# List of supported available CPU Platforms
# https://cloud.google.com/compute/docs/instances/specify-min-cpu-platform#availablezones
class MinCPUPlatform(Enum):
    XEON_ICE_LAKE = "Intel Ice Lake"
    XEON_CASCADE_LAKE = "Intel Cascade Lake"
    XEON_SKYLAKE = "Intel Skylake"
    XEON_BROADWELL = "Intel Broadwell"
    XEON_HASWELL = "Intel Haswell"
    XEON_SANDY_BRIDGE = "Intel Sandy Bridge"
    EPYC_ROME = "AMD Rome"
    EPYC_MILAN = "AMD Milan"


def get_gcp_batch_client(
    sync: bool = True,
) -> Union[batch_v1.BatchServiceClient, batch_v1.BatchServiceAsyncClient]:
    # TODO: Integrate redun version here later.
    client_info = gapic_v1.client_info.ClientInfo(user_agent=f"redun/{version}")
    return (
        batch_v1.BatchServiceClient(client_info=client_info)
        if sync
        else batch_v1.BatchServiceAsyncClient(client_info=client_info)
    )


def gb_to_mib(gb):
    # Convert GiB to MiB.
    return int(gb * 1024)


def batch_submit(
    client: Union[batch_v1.BatchServiceClient, batch_v1.BatchServiceAsyncClient],
    job_name: str,
    project: str,
    region: str,
    machine_type: str,
    vcpus: int,
    memory: int,
    retries: int,
    priority: int,
    max_duration: Optional[str] = None,
    task_count: int = 1,
    mount_buckets: List[str] = [],
    boot_disk_size: Optional[int] = None,
    min_cpu_platform: Optional[MinCPUPlatform] = None,
    accelerators: List[Tuple[str, int]] = [],
    image: Optional[str] = None,
    script: str = "exit 0",
    commands: List[str] = ["exit 0"],
    service_account_email: str = "",
    labels: Dict[str, str] = {},
    **kwargs,  # Ignore extra args
) -> batch_v1.Job:
    # Define what will be done as part of the job.
    runnable = batch_v1.Runnable()

    # Setup volumes
    def make_volume(bucket):
        gcs_bucket = batch_v1.GCS()
        gcs_bucket.remote_path = bucket
        gcs_volume = batch_v1.Volume()
        gcs_volume.gcs = gcs_bucket
        gcs_volume.mount_path = f"/mnt/disks/{bucket}"
        return gcs_volume

    volumes = [make_volume(bucket) for bucket in mount_buckets]

    if image is None:
        runnable.script = batch_v1.Runnable.Script()
        runnable.script.text = script
        # You can also run a script from a file. Just remember, that needs to be a script that's
        # already on the VM that will be running the job. Using runnable.script.text and
        # runnable.script.path is mutually exclusive.
        # runnable.script.path = '/tmp/test.sh'
    else:
        runnable.container = batch_v1.Runnable.Container()
        runnable.container.image_uri = image
        runnable.container.commands = commands
        runnable.container.volumes = [f"{x.mount_path}:{x.mount_path}" for x in volumes]

    task = batch_v1.TaskSpec()
    task.runnables = [runnable]
    task.volumes = volumes

    # We can specify what resources are requested by each task.
    resources = batch_v1.ComputeResource()
    resources.cpu_milli = vcpus * 1000  # in milliseconds per cpu-second.
    # This means the task requires 2 whole CPUs with default value.
    resources.memory_mib = gb_to_mib(memory)
    if boot_disk_size:
        resources.boot_disk_mib = gb_to_mib(boot_disk_size)
    task.compute_resource = resources

    task.max_retry_count = retries

    if max_duration:
        task.max_run_duration = max_duration

    # Tasks are grouped inside a job using TaskGroups.
    # Currently, it's possible to have only one task group.
    group = batch_v1.TaskGroup()
    group.task_count = task_count
    group.task_spec = task

    # Policies are used to define on what kind of virtual machines the tasks will run on.
    # Read more about machine types here: https://cloud.google.com/compute/docs/machine-types
    allocation_policy = batch_v1.AllocationPolicy()
    policy = batch_v1.AllocationPolicy.InstancePolicy()
    policy.machine_type = machine_type
    policy.min_cpu_platform = min_cpu_platform.value if min_cpu_platform else None

    def create_accelerator(typ, count):
        accelerator = batch_v1.AllocationPolicy.Accelerator()
        accelerator.type_ = type
        accelerator.count = count
        return accelerator

    policy.accelerators = list(map(lambda a: create_accelerator(a[0], a[1]), accelerators))

    instances = batch_v1.AllocationPolicy.InstancePolicyOrTemplate()
    if policy.accelerators:
        instances.install_gpu_drivers = True
    instances.policy = policy
    allocation_policy.instances = [instances]

    if service_account_email:
        service_account = batch_v1.ServiceAccount()
        service_account.email = service_account_email
        allocation_policy.service_account = service_account

    job = batch_v1.Job()
    job.priority = priority
    job.task_groups = [group]
    job.allocation_policy = allocation_policy
    job.labels = labels

    # We use Cloud Logging as it's an out of the box available option
    job.logs_policy = batch_v1.LogsPolicy()
    job.logs_policy.destination = batch_v1.LogsPolicy.Destination.CLOUD_LOGGING

    create_request = batch_v1.CreateJobRequest()
    create_request.job = job
    create_request.job_id = job_name
    # The job's parent is the region in which the job will run
    create_request.parent = f"projects/{project}/locations/{region}"

    return client.create_job(create_request)


def list_jobs(
    client: batch_v1.BatchServiceClient, project_id: str, region: str
) -> Iterable[batch_v1.Job]:
    """
    Get a list of all jobs defined in given region.

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        region: name of the region hosting the jobs.

    Returns:
        An iterable collection of Job object.
    """
    return client.list_jobs(parent=f"projects/{project_id}/locations/{region}")


def format_job_name(project_id: str, region: str, job_name: str) -> str:
    """
    Create a job name from provided details

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        region: name of the region hosts the job.
        job_name: fully qualified name of the job you want to retrieve information about.

    Returns:
        A formatted job name.
    """
    return f"projects/{project_id}/locations/{region}/jobs/{job_name}"


def get_job(client: batch_v1.BatchServiceClient, job_name: str) -> batch_v1.Job:
    """
    Retrieve information about a Batch Job.

    Args:
        job_name: fully qualified name of the job you want to retrieve information about.

    Returns:
        A Job object representing the specified job.
    """
    return client.get_job(name=job_name)


def format_task_group_name(project_id: str, region: str, job_name: str, group_name: str) -> str:
    """
    Create a task group name from provided details

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        region: name of the region hosting the jobs.
        job_name: name of the job which tasks you want to list.
        group_name: name of the group of tasks. Usually it's `group0`.

    Return:
        A formatted task group name.
    """
    return f"projects/{project_id}/locations/{region}/jobs/{job_name}/taskGroups/{group_name}"


def list_tasks(client: batch_v1.BatchServiceClient, group_name: str) -> Iterable[batch_v1.Task]:
    """
    Get a list of all jobs defined in given region.

    Args:
        group_name: fully qualified name of the group of tasks you want to look up

    Returns:
        An iterable collection of Task objects.
    """
    return client.list_tasks(parent=group_name)


def format_task_name(
    project_id: str, region: str, job_name: str, group_name: str, task_number: int
) -> str:
    """
    Create a task name from provided details

    Args:
        project_id: project ID or project number of the Cloud project you want to use.
        region: name of the region hosts the job.
        job_name: the name of the job you want to retrieve information about.
        group_name: the name of the group that owns the task you want to check.
            Usually it's `group0`.
        task_number: number of the task you want to look up.

    Return:
        A formatted task name.
    """
    return (
        f"projects/{project_id}/locations/{region}/jobs/{job_name}"
        f"/taskGroups/{group_name}/tasks/{task_number}"
    )


def get_task(client: batch_v1.BatchServiceClient, task_name: str) -> batch_v1.Task:
    """
    Retrieve information about a Task.

    Args:
        task_name: fully qualified name of the task you want to look up.

    Returns:
        A Task object representing the specified task.
    """
    return client.get_task(name=task_name)


def get_gcp_compute_client() -> compute_v1.MachineTypesClient:
    return compute_v1.MachineTypesClient()


@lru_cache(maxsize=None)
def get_compute_machine_type(
    client: compute_v1.MachineTypesClient, project: str, region: str, machine_type: str
) -> compute_v1.types.MachineType:
    """
    Retrieve information about a GCP MachineType

     Args:
         project_id: project ID or project number of the Cloud project you want to use.
         zone: name of the zone for the machine type.
         machine_type: the machine type to get details about.
     Returns:
        A MachineType which provides information about a machines available vCPUs and memory.
    """

    return client.get(project=project, zone=f"{region}-a", machine_type=machine_type)



import logging

log_formatter = logging.Formatter("[redun] %(message)s")
log_handler = logging.StreamHandler()
log_handler.setFormatter(log_formatter)

logger = logging.getLogger("redun")
logger.setLevel(logging.INFO)
logger.addHandler(log_handler)

log_levels = {
    "ERROR": logging.ERROR,
    "WARN": logging.WARN,
    "INFO": logging.INFO,
    "DEBUG": logging.DEBUG,
}



from typing import Optional

from redun import Scheduler, get_context, task
from redun.config import Config
from redun.context import get_context_value
from redun.tests.utils import wait_until


def test_get_context_value() -> None:
    """
    Should be able to fetch values from a context dict.
    """
    context = {
        "foo": {
            "bar": {
                "baz": 10,
                "baaz": True,
            }
        },
        "foo2": "hello",
        "list": [1, 2, 3],
    }
    assert get_context_value(context, "foo.bar.baz", 0) == 10
    assert get_context_value(context, "foo.bar.baaz", False) is True
    assert get_context_value(context, "foo.bar.baz2", 0) == 0
    assert get_context_value(context, "bad", -1) == -1
    assert get_context_value(context, "foo2", "") == "hello"
    assert get_context_value(context, "list", []) == [1, 2, 3]
    assert get_context_value(context, "list.bad", []) == []


def test_context(scheduler: Scheduler) -> None:
    """
    Basic context features.
    """

    task_calls = []

    @task
    def add(a: int, b: int, offset: int = get_context("offset", 3)) -> int:
        task_calls.append("add")
        return a + b + offset

    # The default value of a context variable should be used.
    assert scheduler.run(add(1, 2)) == 6

    # Context can be changed with task_options override.
    # Also caching considers the context arguments.
    assert scheduler.run(add.update_context({"offset": 4})(1, 2)) == 7
    assert task_calls == ["add", "add"]

    # Cache should be used.
    assert scheduler.run(add.update_context({"offset": 4})(1, 2)) == 7
    assert task_calls == ["add", "add"]

    # Alternative update_context call.
    assert scheduler.run(add.update_context(offset=5)(1, 2)) == 8

    @task
    def main():
        return add(3, 4)

    # The parent job's context should be used.
    assert scheduler.run(main.update_context({"offset": 10})()) == 17


def test_context_nested(scheduler: Scheduler) -> None:
    """
    Context should support variable nesting.
    """

    @task
    def task1(x: str = get_context("aa.bb.cc", "foo")) -> str:
        return x

    context = {"aa": {"bb": {"cc": "bar"}}}
    assert scheduler.run(task1.update_context(context)()) == "bar"


def test_context_default_none(scheduler: Scheduler) -> None:
    """
    If default is not provided, it should fallback to None.
    """

    @task
    def task1(x: str = get_context("a.b.c")) -> Optional[str]:
        return x

    assert scheduler.run(task1(), context={"a": {"b": 1}}) is None


def test_context_merge(scheduler: Scheduler) -> None:
    """
    Context should merge from higher tasks.
    """

    @task
    def add(
        a: int,
        b: int,
        offset: int = get_context("offset", 3),
        mult: int = get_context("mult", 1),
    ) -> int:
        return a + b + offset * mult

    # The default value of a context variable should be used.
    assert scheduler.run(add(1, 2)) == 6

    @task
    def main():
        return add(3, 4)

    # The parent job's context should be used.
    assert scheduler.run(main.update_context({"mult": 2})()) == 13


def test_context_run(scheduler: Scheduler) -> None:
    """
    Context should be settable from Scheduler.run().
    """

    @task
    def add(
        a: int,
        b: int,
        offset: int = get_context("offset", 3),
        mult: int = get_context("mult", 1),
    ) -> int:
        return a + b + offset * mult

    assert scheduler.run(add(1, 2), context={"offset": 10, "mult": 20}) == 203


def test_context_config() -> None:
    """
    Context should be settable from Config.
    """

    @task
    def add(
        a: int,
        b: int,
        offset: int = get_context("offset", 3),
        mult: int = get_context("mult", 1),
    ) -> int:
        return a + b + offset * mult

    scheduler = Scheduler(config=Config({"scheduler": {"context": '{"offset": 10, "mult": 20}'}}))
    assert scheduler.run(add(1, 2)) == 203


def test_context_update(scheduler: Scheduler) -> None:
    """
    We should be able to partially update context multiple times.
    """

    @task
    def add(
        a: int,
        b: int,
        offset: int = get_context("offset", 0),
        mult: int = get_context("mult", 1),
    ) -> int:
        return a + b + offset * mult

    assert scheduler.run(add(1, 2)) == 3
    assert scheduler.run(add.update_context({"offset": 4})(1, 2)) == 7
    assert scheduler.run(add.update_context({"offset": 4}).update_context({"mult": 2})(1, 2)) == 11


def test_context_cse(scheduler: Scheduler) -> None:
    """
    Ensure CSE does not collapse jobs with different contexts.
    """

    task_calls = []

    @task
    def add(a, b, offset=get_context("offset", 0)):
        task_calls.append("add")
        return a + b + offset

    @task
    def foo():
        return add(1, 2)

    @task
    def main():
        # The two foo() calls could collapse if context was not considered.
        return foo() + foo.update_context(offset=1)()

    assert scheduler.run(main()) == 7
    assert task_calls == ["add", "add"]


def test_context_cse2(scheduler: Scheduler) -> None:
    """
    Ensure CSE does not collapse jobs with different contexts.
    """

    task_calls = []

    @task
    def add(a, b, offset=get_context("offset", 0)):
        task_calls.append("add")
        return a + b + offset

    @task
    def foo():
        return add(1, 2)

    @task
    def bar():
        # Here we wait for the first foo to finish.
        # This tests the case where we check the db for CSE.
        wait_until(lambda: len(task_calls) > 0)
        return foo.update_context(offset=1)()

    @task
    def main():
        return foo() + bar()

    assert scheduler.run(main()) == 7
    assert task_calls == ["add", "add"]


def test_context_check_valid(scheduler: Scheduler) -> None:
    """
    Ensure check_valid="shallow" cache checking works with changing context.
    """
    task_calls = []

    @task
    def add(a, b, offset=get_context("offset", 0)):
        task_calls.append("add")
        return a + b + offset

    @task(check_valid="shallow")
    def foo():
        return add(1, 2)

    @task
    def bar():
        # Here we wait for the first foo to finish.
        # This tests the case where we check db cache.
        wait_until(lambda: len(task_calls) > 0)
        return foo.update_context(offset=1)()

    @task
    def main():
        # The two foo() calls could collapse if context was not considered.
        return foo() + bar()

    assert scheduler.run(main()) == 7
    assert task_calls == ["add", "add"]



"""Remove length restriction on value type

Revision ID: eb7b95e4e8bf
Revises: cc4f663817b6
Create Date: 2022-11-07 16:22:50.409259

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "eb7b95e4e8bf"
down_revision = "cc4f663817b6"
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name != "sqlite":
        op.alter_column(
            "value",
            "type",
            existing_type=sa.String(length=100),
            type_=sa.String(length=None),
        )


def downgrade():
    if op.get_bind().dialect.name != "sqlite":
        op.alter_column(
            "value",
            "type",
            existing_type=sa.String(length=None),
            type_=sa.String(length=100),
        )



from typing import cast
from unittest.mock import Mock, patch

from redun import File, task
from redun.config import Config
from redun.executors.docker import DockerExecutor
from redun.expression import TaskExpression
from redun.scheduler import Job, Scheduler, Traceback
from redun.tests.utils import use_tempdir
from redun.utils import pickle_dumps


@task
def task1(x: int) -> int:
    return x


@use_tempdir
@patch("redun.executors.docker.run_docker")
@patch("threading.Thread")
@patch("redun.executors.docker.iter_job_status")
@patch.object(Scheduler, "done_job")
@patch.object(Scheduler, "reject_job")
def test_executor_docker(
    reject_job_mock: Mock,
    done_job_mock: Mock,
    iter_job_status_mock: Mock,
    thread_mock: Mock,
    run_docker_mock: Mock,
    scheduler: Scheduler,
) -> None:
    """
    DockerExecutor should run jobs.

    In this test, we patch `threading.Thread` to prevent the monitor thread
    from starting and we call the monitor code ourselves. This removes the
    need to manage multiple threads during the test.
    """

    container_id = "container-1"
    run_docker_mock.return_value = container_id

    config = Config(
        {
            "docker": {
                "image": "my-image",
                "scratch": ".redun_scratch",
            }
        }
    )
    executor = DockerExecutor("docker", scheduler, config=config["docker"])

    # Create and submit a job.
    expr = cast(TaskExpression[int], task1(10))
    job = Job(task1, expr)
    job.eval_hash = "eval_hash"
    job.args = ((10,), {})
    executor.submit(job)

    # Ensure job options were passed correctly to docker.
    scratch_dir = executor._scratch_prefix
    assert executor._code_file
    assert run_docker_mock.call_args[0] == (
        [
            "redun",
            "--check-version",
            ">=0.4.1",
            "oneshot",
            "redun.tests.test_docker_executor",
            "--code",
            executor._code_file.path,
            "--input",
            f"{scratch_dir}/jobs/eval_hash/input",
            "--output",
            f"{scratch_dir}/jobs/eval_hash/output",
            "--error",
            f"{scratch_dir}/jobs/eval_hash/error",
            "task1",
        ],
    )
    assert run_docker_mock.call_args[1] == {
        "image": "my-image",
        "gpus": 0,
        "memory": 4,
        "shared_memory": None,
        "vcpus": 1,
        "volumes": [(scratch_dir, scratch_dir)],
        "interactive": False,
        "include_aws_env": True,
    }

    # Simulate output file created by job.
    output_file = File(f"{scratch_dir}/jobs/eval_hash/output")
    output_file.write(pickle_dumps(task1.func(10)), mode="wb")

    # Simulate container succeeding.
    iter_job_status_mock.return_value = [
        {"jobId": container_id, "status": "SUCCEEDED", "logs": ""},
    ]

    # Manually run monitor logic.
    executor._monitor()

    # Ensure job returns result to scheduler.
    scheduler.done_job.assert_called_with(job, 10)  # type: ignore

    container2_id = "container-2"
    run_docker_mock.return_value = container2_id

    # Create and submit a job.
    expr = cast(TaskExpression[int], task1(11))
    job = Job(task1, expr)
    job.eval_hash = "eval_hash2"
    job.args = ((11,), {})
    executor.submit(job)

    # Simulate the container job failing.
    error = ValueError("Boom")
    error_traceback = Traceback.from_error(error)
    error_file = File(f"{scratch_dir}/jobs/eval_hash2/error")
    error_file.write(pickle_dumps((error, error_traceback)), mode="wb")

    iter_job_status_mock.return_value = [
        {"jobId": container2_id, "status": "FAILED", "logs": ""},
    ]

    # Manually run monitor thread.
    executor._monitor()

    # Ensure job rejected to scheduler.
    scheduler.reject_job.call_args[:2] == (job, error)  # type: ignore



"""add Evaluation model

Revision ID: 647c510a77b1
Revises: 806f5dcb11bf
Create Date: 2020-05-07 16:33:36.216423

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "647c510a77b1"
down_revision = "806f5dcb11bf"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "evaluation",
        sa.Column("eval_hash", sa.String(length=40), nullable=False),
        sa.Column("task_hash", sa.String(length=40), nullable=False),
        sa.Column("args_hash", sa.String(length=40), nullable=False),
        sa.Column("value_hash", sa.String(length=40), nullable=False),
        sa.ForeignKeyConstraint(
            ["task_hash"],
            ["task.hash"],
        ),
        sa.ForeignKeyConstraint(
            ["value_hash"],
            ["value.value_hash"],
        ),
        sa.PrimaryKeyConstraint("eval_hash"),
    )
    op.create_index(op.f("ix_evaluation_task_hash"), "evaluation", ["task_hash"], unique=False)
    op.create_index(op.f("ix_evaluation_value_hash"), "evaluation", ["value_hash"], unique=False)
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_evaluation_value_hash"), table_name="evaluation")
    op.drop_index(op.f("ix_evaluation_task_hash"), table_name="evaluation")
    op.drop_table("evaluation")
    # ### end Alembic commands ###



from typing import Any, Optional, Tuple, Type

from redun.hashing import hash_kwargs, hash_positional_args, hash_struct
from redun.namespace import compute_namespace
from redun.value import Value, get_type_registry


def get_handle_class(handle_class_name: str) -> Type["Handle"]:
    """
    Returns a Handle class from the TypeRegistry.
    """
    klass = get_type_registry().parse_type_name(handle_class_name)
    return klass


def get_fullname(namespace: Optional[str], name: str) -> str:
    """
    Constructs a fullname from a namespace and a name.
    """
    if namespace:
        return namespace + "." + name
    else:
        return name


class Handle(Value):
    """
    A `Value` that opaquely accumulates state as it passes through `Task`s. This subclass of
    `Value` has deep support from the scheduler and its backend, in order to provide a foundation
    for encapsulating interactions with a deterministic, stateful, external system, such as a
    database.

    This model of state supports being initialized, then opaque transformations as `Task`s are
    applied. The hash data and serialization data is designed to be opaque to the actual
    data held within the handle. At initialization time, only the constructor data is captured
    for serialization and hashing. Upon transformation (that is, when a `Handle` is returned by
    a `Task`) the hash is replaced with the hash identifying the `Task` invocation, including
    both code and inputs.

    See the design documents for more discussion of the principles behind the design:
    `docs/source/design.md`.

    In general use, users should not alter the serialization methods.
    """

    type_name = "redun.Handle"

    class HandleInfo:
        """The `fullname` is an identification key for the handle. It is used globally
        across the redun backend to identify other potential instances of this `Handle`.
        However, hashes are used, as usual, to perform equality-like tests.
        """

        def __init__(
            self,
            name: str,
            args: Tuple,
            kwargs: dict,
            class_name: str,
            namespace: Optional[str] = None,
            call_hash: str = "",
            hash: Optional[str] = None,
            key: str = "",
        ):
            # The name and namespace comprise the fullname.
            self.name = name
            self.namespace = namespace
            self.fullname = get_fullname(self.namespace, self.name)

            # The args and kwargs are captured for reconstruction after serialization.
            self.args = args
            self.kwargs = kwargs
            self.class_name = class_name

            # After transformation of a handle, or "calls", the hash of the invocation is used
            # instead of the initialization data.
            self.call_hash = call_hash

            # The key contributes to the hash and allows us to differentiate between
            # otherwise-identical handles.
            self.key = key

            # For a newly created object, we have to compute the hash. However, we include the hash
            # in our serialization output, so this will be returned to us when (re)constructing a
            # deserialized object.
            self.hash = hash or self.get_hash()

            # Cache of recent parent handles used by redun to aid recording.
            self.is_recorded = False  # TODO: Would be nice to move this to backend.
            self.fork_parent: Optional[Handle] = None

        def get_state(self) -> dict:
            """
            Returns serializable state dict.
            """
            return {
                "name": self.name,
                "namespace": self.namespace,
                "args": self.args,
                "kwargs": self.kwargs,
                "class_name": self.class_name,
                "call_hash": self.call_hash,
                "key": self.key,
                "hash": self.hash,
            }

        def get_hash(self) -> str:
            """
            Returns hash of the handle.

            Implementation note: the Handle state model requires that the computation proceeds
            differently for new Handles and ones that are computed by other tasks.
            """
            if self.call_hash:
                # For a Handle that is the output of a Task, switch to using information about the
                # Task invocation.
                return hash_struct(
                    ["Handle", self.fullname, "call_hash", self.key, self.call_hash]
                )
            else:
                # The initial state is defined by constructor data only.
                args_hash = hash_positional_args(get_type_registry(), self.args)
                kwargs_hash = hash_kwargs(get_type_registry(), self.kwargs)
                return hash_struct(
                    ["Handle", self.fullname, "init", self.key, args_hash, kwargs_hash]
                )

        def update_hash(self) -> None:
            self.hash = self.get_hash()

        def apply_call(self, handle: "Handle", call_hash: str) -> "Handle":
            handle2 = self.clone(handle)
            handle2.__handle__.call_hash = call_hash
            handle2.__handle__.key = ""
            handle2.__handle__.update_hash()
            return handle2

        def fork(self, handle: "Handle", key: str) -> "Handle":
            handle2 = self.clone(handle)

            # Note: When forking, the previous Handle hash is used as the call_hash.
            handle2.__handle__.call_hash = handle.__handle__.hash
            handle2.__handle__.key = key
            handle2.__handle__.update_hash()
            handle2.__handle__.fork_parent = handle
            return handle2

        def clone(self, handle: "Handle") -> "Handle":
            # Create new handle instance.
            klass = get_handle_class(self.class_name)
            handle2 = klass.__new__(
                klass, self.name, namespace=self.namespace, *self.args, **self.kwargs
            )

            # Copy over attributes to new handle.
            ignore_attrs = {"__handle__"}
            for key, value in handle.__dict__.items():
                if key not in ignore_attrs:
                    handle2.__dict__[key] = value

            return handle2

    def __new__(
        cls,
        name: Optional[str] = None,
        *args,
        namespace: Optional[str] = None,
        **kwargs,
    ) -> "Handle":
        handle = super().__new__(cls)

        # Note: name is None when loading from a pickle.
        if name is not None:
            # Set HandleInfo within __new__ so that user can't forget.
            handle.__handle__ = cls.HandleInfo(
                name=name,
                namespace=compute_namespace(cls, namespace),
                args=args,
                kwargs=kwargs,
                class_name=cls.type_name,
            )
        return handle

    def __repr__(self) -> str:
        return "{class_name}(fullname={fullname}, hash={hash})".format(
            class_name=self.__handle__.class_name,
            fullname=self.__handle__.fullname,
            hash=(self.__handle__.hash or "")[:8],
        )

    def __getattr__(self, attr: str) -> Any:
        """
        Proxy attribute access to `self.instance` if it exists.
        """
        if "instance" in self.__dict__:
            return getattr(self.instance, attr)
        else:
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(type(self).__name__, attr)
            )

    def __getstate__(self) -> dict:
        """
        Returns dict for serialization.
        """
        return self.__handle__.get_state()

    def __setstate__(self, state: dict) -> None:
        """
        Sets state from dict for deserialization.
        """
        self.__handle__ = self.HandleInfo(**state)
        self.__handle__.is_recorded = True
        self.__init__(state["name"], *state["args"], **state["kwargs"])  # type: ignore

    def apply_call(self, call_hash: str) -> "Handle":
        """
        Returns a new Handle derived from this one assuming passage through a call with call_hash.
        """
        return self.__handle__.apply_call(self, call_hash)

    def fork(self, key: str) -> "Handle":
        """
        Forks the handle into a second one for use in parallel tasks. A key must be provided
        to differentiate the fork from the original (although the original may have a key)
        """
        return self.__handle__.fork(self, key)

    def is_valid(self) -> bool:
        """
        Returns True if handle is still valid (i.e., has not been rolled back).
        """
        from redun.scheduler import get_current_scheduler

        if self.type_name != self.__handle__.class_name:
            # Handle class_name might be out of date from deserialization.
            return False

        scheduler = get_current_scheduler()
        assert scheduler
        return scheduler.backend.is_valid_handle(self)

    def get_hash(self, data: Optional[bytes] = None) -> str:
        """
        Returns a hash of the handle.
        """
        return self.__handle__.hash

    def preprocess(self, preprocess_args: dict) -> "Handle":
        """
        Forks a handle as it passes into a task.
        """
        call_order = preprocess_args["call_order"]
        return self.fork(self.__handle__.key or str(call_order))

    def postprocess(self, postprocess_args: dict) -> "Handle":
        """
        Applies the call_hash to the handle as it returns from a task.
        """
        return self.apply_call(postprocess_args["pre_call_hash"])



from typing import cast

import pytest
from sqlalchemy.orm import Session

from redun import File, Scheduler, task
from redun.backends.db.query import CallGraphQuery, Job, infer_id
from redun.scheduler import scheduler_task
from redun.tests.utils import use_tempdir


def test_infer_id(scheduler: Scheduler, session: Session) -> None:
    """
    infer_id should be able to detect specialty ids (-, ~2, etc).
    """

    @task
    def add(x: int, y: int) -> int:
        return x + y

    # Perform some Executions.
    assert scheduler.run(add(1, 2)) == 3
    assert scheduler.run(add(1, 3)) == 4
    assert scheduler.run(add(1, 4)) == 5

    # Get most recent job.
    exec1 = infer_id(session, "-")
    assert exec1.job.call_node.value_parsed == 5

    exec1 = infer_id(session, "~1")
    assert exec1.job.call_node.value_parsed == 5

    # Get second most recent job.
    exec2 = infer_id(session, "~2")
    assert exec2.job.call_node.value_parsed == 4

    # Get third most recent job.
    exec3 = infer_id(session, "~3")
    assert exec3.job.call_node.value_parsed == 3

    # Infer id by prefix.
    assert infer_id(session, exec1.id[:8]) == exec1


@use_tempdir
def test_infer_id_file(scheduler: Scheduler, session: Session) -> None:
    """
    infer_id should be able to detect filenames.
    """

    @task
    def write_file(name: str) -> File:
        file = File(name)
        file.write("hello")
        return file

    @task
    def read_file(file: File) -> str:
        return cast(str, file.read())

    # Perform some Executions.
    scheduler.run(write_file("file1"))

    file = File("file2")
    file.write("bye")
    scheduler.run(read_file(file))

    # Get file info.
    file_info = infer_id(session, "file1")
    assert file_info[0].path == "file1"
    assert isinstance(file_info[1], Job)
    assert file_info[1].call_node.value_parsed.path == "file1"
    assert file_info[2] == "result"

    file_info = infer_id(session, "file2")
    assert file_info[0].path == "file2"
    assert isinstance(file_info[1], Job)
    assert file_info[1].call_node.arguments[0].value_parsed.path == "file2"
    assert file_info[2] == "arg"


def test_status(scheduler: Scheduler, session: Session) -> None:
    """
    We should be able to filter on Execution and Job status correctly.
    """

    @task
    def div(x: float, y: float) -> float:
        return x / y

    @scheduler_task()
    def kill(scheduler, parent_job, expr):
        # Immediately kill the workflow. This skips recording end_time of the active jobs, so they
        # will be considered RUNNING.
        scheduler.reject_job(None, Exception())

    @task
    def do_kill():
        return kill()

    # Perform some Executions.
    assert scheduler.run(div(4, 2)) == 2
    assert scheduler.run(div(4, 2)) == 2
    assert scheduler.run(div(8, 2)) == 4

    with pytest.raises(ZeroDivisionError):
        assert scheduler.run(div(1, 0))

    with pytest.raises(Exception):
        assert scheduler.run(do_kill())

    query = CallGraphQuery(session)

    # Assert job statuses are correct.
    assert len(list(query.filter_job_statuses(["DONE"]).all())) == 2
    assert len(list(query.filter_job_statuses(["CACHED"]).all())) == 1
    assert len(list(query.filter_job_statuses(["FAILED"]).all())) == 1
    assert len(list(query.filter_job_statuses(["RUNNING"]).all())) == 1

    # Assert execution statuses are correct.
    assert len(list(query.filter_execution_statuses(["DONE"]).all())) == 3
    assert len(list(query.filter_execution_statuses(["FAILED"]).all())) == 1
    assert len(list(query.filter_execution_statuses(["RUNNING"]).all())) == 1



import argparse
import os
import re
from functools import reduce
from itertools import islice
from typing import Any, Callable, Dict, Iterable, Iterator, List, Optional, Set, Tuple

import sqlalchemy as sa
from sqlalchemy.orm import Session
from sqlalchemy.orm.query import Query
from sqlalchemy.sql.expression import cast as sa_cast
from sqlalchemy.sql.expression import select

from redun.backends.db import (
    JSON,
    Argument,
    Base,
    CallNode,
    Execution,
    File,
    Job,
    Subvalue,
    Tag,
    Task,
    Value,
)
from redun.tags import ANY_VALUE, parse_tag_key_value

REDUN_ERROR_TYPE_NAME = "redun.ErrorValue"


def setup_query_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """
    Add actions to an ArgumentParser to support querying a CallGraph.
    """

    # Log command.
    parser.add_argument(
        "--all",
        action="store_true",
        help="Show all record types (Execution, Job, Task, CallNode, Value).",
    )
    parser.add_argument("--exec", action="store_true", help="Show related executions.")
    parser.add_argument("--job", action="store_true", help="Show related jobs.")
    parser.add_argument("--call-node", action="store_true", help="Show related CallNodes.")
    parser.add_argument("--task", action="store_true", help="Show related tasks.")
    parser.add_argument("--value", action="store_true", help="Show related values.")
    parser.add_argument("--file", action="store_true", help="Show related files.")
    parser.add_argument(
        "--exec-status",
        help="Filter executions by status (comma separated: RUNNING,FAILED,DONE).",
    )
    parser.add_argument(
        "--job-status",
        help="Filter jobs by status (comma separated: RUNNING,FAILED,CACHED,DONE).",
    )
    parser.add_argument("--task-name", action="append", help="Filter tasks and jobs by task name.")
    parser.add_argument("--task-hash", action="append", help="Filter tasks and jobs by task hash.")
    parser.add_argument("--value-type", action="append", help="Filter Values by their type.")
    parser.add_argument("--file-path", action="append", help="Filter by File path.")
    parser.add_argument("--exec-id", help="Filter by execution ids (comma separated ids).")
    parser.add_argument(
        "--result", action="append", help="Filter jobs by their result value hashes."
    )
    parser.add_argument(
        "--arg", action="append", help="Filter jobs by their argument value hashes."
    )
    parser.add_argument("-t", "--tag", action="append", help="Filter by tag (format: key=value).")
    parser.add_argument(
        "--exec-tag",
        action="append",
        help="Filter by execution tag (format: key=value).",
    )
    return parser


def parse_callgraph_query(query: "CallGraphQuery", args: argparse.Namespace) -> "CallGraphQuery":
    """
    Parse an argparse.Namespace into a CallGraphQuery.
    """

    record_types = set()

    # Filter by execution status.
    if args.exec_status:
        record_types.add("Execution")
        query = query.filter_execution_statuses(args.exec_status.split(","))

    # Filter by job status.
    if args.job_status:
        record_types.add("Job")
        query = query.filter_job_statuses(args.job_status.split(","))

    # Filter by value types.
    if args.value_type:
        # Implies filtering by Value type.
        record_types.add("Value")
        query = query.filter_value_types(args.value_type)
    if args.file_path:
        # Implies filtering by Value type.
        record_types.add("Value")
        query = query.filter_file_paths(args.file_path)

    # Filter by record type.
    if args.all:
        record_types.update(CallGraphQuery.MODEL_NAMES)
    if args.exec:
        record_types.add("Execution")
    if args.job:
        record_types.add("Job")
    if args.call_node:
        record_types.add("CallNode")
    if args.task:
        record_types.add("Task")
    if args.value:
        record_types.add("Value")
    if args.file:
        record_types = {"Value"}
        query = query.filter_value_types(["redun.File"])
    if record_types:
        query = query.filter_types(record_types)

    # Filter by execution id.
    if args.exec_id:
        query = query.filter_execution_ids(args.exec_id.split(","))

    # Filter task properties.
    if args.task_name:
        query = query.filter_task_names(args.task_name)
    if args.task_hash:
        query = query.filter_task_hashes(args.task_hash)

    if args.result:
        query = query.filter_results(args.result)
    if args.arg:
        query = query.filter_arguments(args.arg)

    # Filter by tags.
    if args.tag:
        tags = [parse_tag_key_value(tag, value_required=False) for tag in args.tag]
        query = query.filter_tags(tags)
    if args.exec_tag:
        tags = [parse_tag_key_value(tag, value_required=False) for tag in args.exec_tag]
        query = query.filter_execution_tags(tags)

    return query


class CallGraphQuery:
    """
    Query class for efficiently querying across all models in a CallGraph.
    """

    # Look up for model primary keys.
    MODEL_PKS = {
        Execution: "id",
        Job: "id",
        CallNode: "call_hash",
        Value: "value_hash",
        Task: "hash",
    }
    MODEL_CLASSES = {
        "Execution": Execution,
        "Job": Job,
        "CallNode": CallNode,
        "Task": Task,
        "Value": Value,
    }
    MODEL_NAMES = list(MODEL_CLASSES.keys())

    ExecTag = sa.orm.aliased(Tag)

    def __init__(
        self,
        session: sa.orm.Session,
        joins: Optional[Set[str]] = None,
        execution_joins: Optional[List[Callable[[Query], Query]]] = None,
        filters: Optional[List] = None,
        order_by: Optional[str] = None,
        filter_types: Optional[Set] = None,
        executions: Optional[Query] = None,
        jobs: Optional[Query] = None,
        call_nodes: Optional[Query] = None,
        tasks: Optional[Query] = None,
        values: Optional[Query] = None,
        value_subqueries: Optional[List[Query]] = None,
    ):
        self._session = session
        self._joins: Set[str] = joins or set()
        self._execution_joins: List[Callable[[Query], Query]] = execution_joins or []
        self._filters: List = filters or []
        self._order_by = order_by
        self._filter_types = filter_types if filter_types is not None else set(self.MODEL_NAMES)

        # Subqueries.
        self._executions = executions or self._session.query(Execution)
        self._jobs = jobs or self._session.query(Job)
        self._call_nodes = call_nodes or self._session.query(CallNode)
        self._tasks = tasks or self._session.query(Task)
        self._values = values if values is not None else self._session.query(Value)
        self._value_subqueries = value_subqueries

    @property
    def subqueries(self) -> Iterable[Tuple[str, Query]]:
        """
        Iterates through all subqueries.

        Yields
        ------
        (type_name, subquery) : Tuple[str, Query]
        """
        return [
            ("Execution", self._executions),
            ("Job", self._jobs),
            ("CallNode", self._call_nodes),
            ("Task", self._tasks),
            ("Value", self._values),
        ]

    def clone(self, **kwargs: Any) -> "CallGraphQuery":
        """
        Returns a clone of the query with updates specified by `kwargs`.
        """
        clone_kwargs: Dict[str, Any] = {
            "joins": self._joins,
            "execution_joins": self._execution_joins,
            "filters": self._filters,
            "order_by": self._order_by,
            "filter_types": self._filter_types,
            "executions": self._executions,
            "jobs": self._jobs,
            "call_nodes": self._call_nodes,
            "tasks": self._tasks,
            "values": self._values,
            "value_subqueries": self._value_subqueries,
        }
        clone_kwargs.update(kwargs)
        return CallGraphQuery(self._session, **clone_kwargs)

    def filter_types(self, types: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by record type.
        """
        return self.clone(filter_types=set(types))

    def filter_ids(self, _ids: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by record ids.
        """
        ids = list(_ids)

        def filter(query):
            return query.clone(
                executions=query._executions.filter(Execution.id.in_(ids)),
                jobs=query._jobs.filter(Job.id.in_(ids)),
                call_nodes=query._call_nodes.filter(CallNode.call_hash.in_(ids)),
                tasks=query._tasks.filter(Task.hash.in_(ids)),
                values=query._values.filter(Value.value_hash.in_(ids)),
            )

        return self.clone(filters=self._filters + [filter])

    def like_id(self, id: str) -> "CallGraphQuery":
        """
        Filter query by record id prefix `id`.
        """

        def filter(query):
            pattern = id + "%"
            return query.clone(
                executions=query._executions.filter(Execution.id.like(pattern)),
                jobs=query._jobs.filter(Job.id.like(pattern)),
                call_nodes=query._call_nodes.filter(CallNode.call_hash.like(pattern)),
                tasks=(
                    self._tasks.filter(Task.hash.like(pattern)).union(
                        self._tasks.filter(Task.name == id)
                    )
                ),
                values=query._values.filter(Value.value_hash.like(pattern)),
            )

        return self.clone(filters=self._filters + [filter])

    def _join_files(self):
        return self.clone(values=self._values.join(File, File.value_hash == Value.value_hash))

    def _join_jobs(self):
        return self.clone(executions=self._executions.join(Job, Job.id == Execution.job_id))

    def _join_tasks(self):
        return self.clone(
            jobs=self._jobs.join(Task, Task.hash == Job.task_hash),
        )

    def _join_values(self):
        return self.clone(
            executions=self._executions.outerjoin(
                CallNode, CallNode.call_hash == Job.call_hash
            ).outerjoin(Value, Value.value_hash == CallNode.value_hash),
            jobs=self._jobs.outerjoin(CallNode, CallNode.call_hash == Job.call_hash).outerjoin(
                Value, Value.value_hash == CallNode.value_hash
            ),
        )

    def _join_executions(self):
        """
        Join Executions to each subquery.
        """
        query = self

        # Values can be joined to Execution via four possible paths. Thus far,
        # we have found it most efficient to perform all four paths and union
        # their results.

        # Values can connect to CallNodes either directly as a result or
        # through Argument as an argument. Values can also be a subvalue of a
        # parent Value. These possibilities combine to give four join paths.
        value_results = query._values.join(CallNode, CallNode.value_hash == Value.value_hash).join(
            Job, Job.call_hash == CallNode.call_hash
        )
        value_args = (
            query._values.join(Argument, Argument.value_hash == Value.value_hash)
            .join(CallNode, CallNode.call_hash == Argument.call_hash)
            .join(Job, Job.call_hash == CallNode.call_hash)
        )
        subvalue_results = (
            query._values.join(Subvalue, Subvalue.value_hash == Value.value_hash)
            .join(CallNode, CallNode.value_hash == Subvalue.parent_value_hash)
            .join(Job, Job.call_hash == CallNode.call_hash)
        )
        subvalue_args = (
            query._values.join(Subvalue, Subvalue.value_hash == Value.value_hash)
            .join(Argument, Argument.value_hash == Subvalue.parent_value_hash)
            .join(CallNode, CallNode.call_hash == Argument.call_hash)
            .join(Job, Job.call_hash == CallNode.call_hash)
        )
        subqueries = [value_results, value_args, subvalue_results, subvalue_args]

        # Perform additional joins on execution.
        for i, _ in enumerate(subqueries):
            for join in self._execution_joins:
                subqueries[i] = join(subqueries[i])

        return query.clone(
            executions=query._executions,
            jobs=query._jobs,
            call_nodes=(query._call_nodes.join(Job, Job.call_hash == CallNode.call_hash)),
            tasks=(query._tasks.join(Job, Job.task_hash == Task.hash)),
            value_subqueries=subqueries,
        )

    def filter_execution_ids(self, execution_ids: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by Execution ids.
        """
        execution_ids = list(execution_ids)

        def filter(query):
            return query.clone(
                executions=query._executions.filter(Execution.id.in_(execution_ids)),
                jobs=query._jobs.filter(Job.execution_id.in_(execution_ids)),
                call_nodes=query._call_nodes.filter(Job.execution_id.in_(execution_ids)),
                tasks=query._tasks.filter(Job.execution_id.in_(execution_ids)),
                values=query._values.filter(Job.execution_id.in_(execution_ids)),
            )

        return self.clone(
            joins=self._joins | {"execution"},
            filters=self._filters + [filter],
        )

    def _job_status_term(self, status: str) -> Any:
        """
        Returns SQLAlchemy filter term for finding Jobs with status `status`.
        """
        if status == "RUNNING":
            return Job.end_time.is_(None) & Job.call_hash.is_(None)
        elif status == "CACHED":
            return Job.cached.is_(True)
        elif status == "FAILED":
            return Value.type == REDUN_ERROR_TYPE_NAME
        elif status == "DONE":
            return Job.cached.is_(False) & (Value.type != REDUN_ERROR_TYPE_NAME)
        else:
            raise NotImplementedError(status)

    def filter_execution_statuses(self, execution_statuses: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Execution status.
        """
        assert execution_statuses

        # Convert execution statuses into root job statuses.
        # When the root job is CACHED, we consider the Execution DONE.
        job_statuses = list(execution_statuses)
        if "DONE" in job_statuses:
            job_statuses.append("CACHED")

        execution_clause = reduce(sa.or_, map(self._job_status_term, job_statuses))

        def filter(query):
            return query.clone(executions=query._executions.filter(execution_clause))

        return self.clone(
            filter_types=self._filter_types & {"Execution"},
            joins=self._joins | {"job", "value"},
            filters=self._filters + [filter],
            order_by="time",
        )

    def filter_job_statuses(self, job_statuses: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Job status.
        """
        assert job_statuses

        job_clause = reduce(sa.or_, map(self._job_status_term, job_statuses))

        def filter(query):
            return query.clone(jobs=query._jobs.filter(job_clause))

        return self.clone(
            filter_types=self._filter_types & {"Job"},
            filters=self._filters + [filter],
            joins=self._joins | {"value"},
        )

    def filter_task_names(self, task_names: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Task names.
        """

        def task_term(task_name: str):
            if "." in task_name:
                namespace, name = task_name.rsplit(".", 1)
                return (Task.namespace == namespace) & (Task.name == name)
            else:
                return Task.name == task_name

        clause = reduce(sa.or_, map(task_term, task_names))

        def filter(query):
            return query.clone(
                jobs=query._jobs.filter(clause),
                tasks=query._tasks.filter(clause),
            )

        return self.clone(
            filter_types=self._filter_types & {"Job", "Task"},
            joins=self._joins | {"task"},
            filters=self._filters + [filter],
        )

    def filter_task_hashes(self, task_hashes: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by Task hashes.
        """
        return self.clone(
            filter_types=self._filter_types & {"Job", "Task"},
            jobs=self._jobs.filter(Job.task_hash.in_(task_hashes)),
            tasks=self._tasks.filter(Task.hash.in_(task_hashes)),
        )

    def filter_arguments(self, value_hashes: List[str]) -> "CallGraphQuery":
        """
        Filter jobs by argument values.
        """
        jobs = (
            self._jobs.join(CallNode, Job.call_hash == CallNode.call_hash)
            .join(Argument, Argument.call_hash == CallNode.call_hash)
            .filter(Argument.value_hash.in_(value_hashes))
        ).union(
            self._jobs.join(CallNode, Job.call_hash == CallNode.call_hash)
            .join(Argument, Argument.call_hash == CallNode.call_hash)
            .join(Subvalue, Subvalue.parent_value_hash == Argument.value_hash)
            .filter(Subvalue.value_hash.in_(value_hashes))
        )
        return self.clone(filter_types=self._filter_types & {"Job"}, jobs=jobs)

    def filter_results(self, value_hashes: List[str]) -> "CallGraphQuery":
        """
        Filter jobs by result values.
        """
        jobs = (
            self._jobs.join(CallNode, Job.call_hash == CallNode.call_hash).filter(
                CallNode.value_hash.in_(value_hashes)
            )
        ).union(
            self._jobs.join(CallNode, Job.call_hash == CallNode.call_hash)
            .join(Subvalue, Subvalue.parent_value_hash == CallNode.value_hash)
            .filter(Subvalue.value_hash.in_(value_hashes))
        )
        return self.clone(filter_types=self._filter_types & {"Job"}, jobs=jobs)

    def filter_value_types(self, value_types: Iterable[str]) -> "CallGraphQuery":
        """
        Filter query by Value types.
        """

        def filter(query):
            return query.clone(values=query._values.filter(Value.type.in_(value_types)))

        return self.clone(
            filter_types=self._filter_types & {"Value"},
            filters=self._filters + [filter],
        )

    def filter_file_paths(self, paths: Iterable[str]) -> "CallGraphQuery":
        """
        Filter by File.path patterns.

        `paths` can contain "*" to perform wildcard matching.
        """
        # Convert path patterns into db LIKE patterns.
        like_patterns = [path.replace("*", "%") for path in paths]

        def filter(query):
            return query.clone(
                values=query._values.filter(
                    sa.or_(File.path.like(like_pattern) for like_pattern in like_patterns)
                )
            )

        return self.clone(
            filter_types=self._filter_types & {"Value"},
            joins=self._joins | {"file"},
            filters=self._filters + [filter],
        )

    def _query_filter_tags(
        self,
        query: Query,
        entity_id_col: Any,
        table: Any,
        tags: Iterable[Tuple[str, Any]],
    ) -> Query:
        """
        Build query for filtering tags.
        """
        for key, value in tags:
            tag_query = self._session.query(table).filter(
                table.is_current.is_(True), table.key == key
            )
            if value is not ANY_VALUE:
                tag_query = tag_query.filter(table.value == sa_cast(value, JSON))
            # We perform this filter as a subquery in order to find entities
            # with multiple matching tags.
            tag_query = tag_query.subquery()
            query = query.join(tag_query, tag_query.c.entity_id == entity_id_col)
        return query

    def filter_tags(self, tags: Iterable[Tuple[str, Any]]) -> "CallGraphQuery":
        """
        Filter by tags.
        """

        def filter(query):
            return query.clone(
                executions=self._query_filter_tags(query._executions, Execution.id, Tag, tags),
                jobs=self._query_filter_tags(query._jobs, Job.id, Tag, tags),
                call_nodes=self._query_filter_tags(
                    query._call_nodes, CallNode.call_hash, Tag, tags
                ),
                tasks=self._query_filter_tags(query._tasks, Task.hash, Tag, tags),
                values=self._query_filter_tags(query._values, Value.value_hash, Tag, tags),
            )

        return self.clone(
            filters=self._filters + [filter],
        )

    def filter_execution_tags(self, tags: Iterable[Tuple[str, Any]]) -> "CallGraphQuery":
        """
        Filter by tag on executions.
        """

        def filter(query):
            # Values will be filtered by the execution_join below.
            return query.clone(
                executions=self._query_filter_tags(
                    query._executions, Execution.id, self.ExecTag, tags
                ),
                jobs=self._query_filter_tags(query._jobs, Job.execution_id, self.ExecTag, tags),
                call_nodes=self._query_filter_tags(
                    query._call_nodes, Job.execution_id, self.ExecTag, tags
                ),
                tasks=self._query_filter_tags(query._tasks, Job.execution_id, self.ExecTag, tags),
            )

        def exec_join(sa_query: Query) -> Query:
            return self._query_filter_tags(sa_query, Job.execution_id, self.ExecTag, tags)

        return self.clone(
            joins=self._joins | {"execution"},
            execution_joins=self._execution_joins + [exec_join],
            filters=self._filters + [filter],
        )

    def order_by(self, order_by: str) -> "CallGraphQuery":
        """
        Order query.

        order_by: str
            The only supported value is 'time'.
        """
        return self.clone(order_by=order_by)

    def build(self):
        """
        Apply all joins and filters to subqueries.
        """
        query = self

        # Perform joins.
        if "file" in self._joins:
            query = query._join_files()
        if "job" in self._joins:
            query = query._join_jobs()
        if "value" in self._joins:
            query = query._join_values()
        if "task" in self._joins:
            query = query._join_tasks()
        if "execution" in self._joins:
            query = query._join_executions()

        # Perform filters.
        for filter in self._filters:
            query = filter(query)

        # Filter and union value subqueries.
        if query._value_subqueries:
            subqueries = [self.clone(values=subquery) for subquery in query._value_subqueries]
            for i, _ in enumerate(subqueries):
                for filter in self._filters:
                    subqueries[i] = filter(subqueries[i])
            values = subqueries[0]._values.union(*(q._values for q in subqueries[1:]))
            query = query.clone(values=values)

        # Perform order_by.
        if self._order_by == "time":
            Job2 = sa.orm.aliased(Job)
            query = query.clone(
                executions=(
                    query._executions.add_columns(Job2.start_time)
                    .join(Job2, Execution.job_id == Job2.id)
                    .order_by(Job2.start_time.desc())
                ),
                jobs=(query._jobs.order_by(Job.start_time.desc())),
            )
        elif self._order_by:
            raise NotImplementedError(self._order_by)

        # Remove filters, joins, etc since they are now built in.
        return query.clone(filters=[], joins={}, order_by=None)

    def empty(self) -> "CallGraphQuery":
        """
        Returns an empty query.
        """
        query = CallGraphQuery(self._session)
        return self.clone(
            executions=query._executions.filter(False),
            jobs=query._jobs.filter(False),
            call_nodes=query._call_nodes.filter(False),
            tasks=query._tasks.filter(False),
            values=query._values.filter(False),
        )

    def all(self) -> Iterator[Base]:
        """
        Yields all records matching query.
        """
        query = self.build()

        for record_type, subquery in query.subqueries:
            if record_type in self._filter_types:
                # Normally, `subquery.statement.distinct()` would be enough here
                # but the extra select() wrapper is needed when using unions.
                # Also, select() doesn't accept string names for now, so we need to look up
                # the actual type.
                selectable = select(self.MODEL_CLASSES[record_type]).from_statement(
                    subquery.statement.distinct()
                )
                yield from self._session.execute(selectable).scalars()

    def one(self) -> Base:
        """
        Returns exactly one record. Raises error if too few or too many.
        """
        [result] = self.all()
        return result

    def one_or_none(self) -> Base:
        """
        Returns exactly one record. Returns None if too few or too many.
        """
        results = list(self.all())
        if not results:
            return None
        [result] = results
        return result

    def first(self) -> Optional[Base]:
        """
        Returns first record if it exists.
        """
        return next(self.all(), None)

    def limit(self, size) -> Iterator[Base]:
        """
        Yields at most `size` records from query.
        """
        query = self.build().clone(
            executions=self._executions.limit(size),
            jobs=self._jobs.limit(size),
            call_nodes=self._call_nodes.limit(size),
            tasks=self._tasks.limit(size),
            values=self._values.limit(size),
        )
        yield from islice(query.all(), 0, size)

    def page(self, page: int, page_size: int) -> Iterator[Base]:
        """
        Yields a page-worth of results from query.

        page is zero-indexed.
        """
        offset = page * page_size

        record_type2subquery = dict(self.build().subqueries)
        returned = 0
        seen = 0
        for record_type, count in self.count():
            subquery = record_type2subquery[record_type]

            seen += count
            if seen <= offset:
                continue

            remainder = offset - (seen - count)
            if remainder > 0:
                subquery = subquery.offset(remainder)

            for record in subquery.limit(page_size - returned):
                returned += 1
                yield record

    def count(self) -> Iterator[Tuple[str, int]]:
        """
        Returns counts for each record type.
        """
        query = self.build()
        for record_type, subquery in query.subqueries:
            if record_type in self._filter_types:
                yield record_type, subquery.distinct().count()

    def select(self, *columns: Iterable[str], flat: bool = False) -> Iterator[Any]:
        """
        Select columns to return in query.
        """
        for record in self.all():
            row = []
            for column in columns:
                if column == "id":
                    pk = self.MODEL_PKS[type(record)]
                    row.append(getattr(record, pk))

                elif column == "type":
                    row.append(type(record).__name__)

                else:
                    raise ValueError(f"Unknown column '{column}'.")

            if flat:
                [value] = row
                yield value
            else:
                yield tuple(row)


def infer_id(session: Session, id: str, include_files: bool = True, required: bool = False) -> Any:
    """
    Try to infer the record based on an id prefix.
    """

    record = infer_specialty_id(session, id, include_files=include_files)
    if record:
        return record

    query = CallGraphQuery(session).like_id(id)
    if required:
        return query.one()
    else:
        return query.first()


def infer_specialty_id(session: Session, id: str, include_files: bool = True) -> Any:
    """
    Try to infer the record based on speciality id (e.g. file paths, `-` shorthand).
    """

    if include_files and (os.path.exists(id) or ":" in id):
        # Looks like a local path.
        file_info = find_file(session, id)
        if file_info:
            return file_info

    # Determine if id is an Execution shorthand.
    executions = (
        session.query(Execution)
        .join(Job, Job.id == Execution.job_id)
        .order_by(Job.start_time.desc())
    )

    if id == "-":
        # Most recent execution.
        return executions.first()

    elif re.match(r"^~\d+", id):
        # Recent execution by index ~1, ~2, etc.
        index = int(id[1:])
        if index == 0:
            raise ValueError("Execution index should >= 1.")
        try:
            return executions[index - 1]
        except IndexError:
            raise ValueError(f"Execution index too high: {index}")

    return None


def find_file(session: Session, path: str) -> Optional[Tuple[File, Job, str]]:
    """
    Find a File by its path.

    - Prefer instance of File as result from a Task.
      - Amongst result, prefer the most recent.
    - Otherwise, search for File as argument to a Task.
      - Amongst arguments, prefer the most recent.

    - For both results and arguments, also search whether File was a Subvalue
      (e.g. a value within a list, dict, etc).

    - We prefer the most recent, since it has the best chance of still being valid.
    """

    # Search for File as a Value resulting from a Task.
    row = (
        session.query(File, Job)
        .join(CallNode, CallNode.value_hash == File.value_hash)
        .join(Job, CallNode.call_hash == Job.call_hash)
        .filter(File.path == path)
        .order_by(Job.end_time.desc())
        .first()
    )

    # Search for File as a Subvalue resulting from a Task.
    row2 = (
        session.query(File, Job)
        .join(Subvalue, File.value_hash == Subvalue.value_hash)
        .join(CallNode, Subvalue.parent_value_hash == CallNode.value_hash)
        .join(Job, CallNode.call_hash == Job.call_hash)
        .filter(File.path == path)
        .order_by(Job.end_time.desc())
        .first()
    )

    if row or row2:
        # Return the most recent file and job reference.
        if not row:
            return (row2[0], row2[1], "result")
        elif not row2:
            return (row[0], row[1], "result")
        else:
            _, job = row
            _, job2 = row2
            if job.end_time > job2.end_time:
                return (row[0], row[1], "result")
            else:
                return (row2[0], row2[1], "result")

    # Search for File as a Argument to a Task.
    row3 = (
        session.query(File, Job)
        .join(Argument, File.value_hash == Argument.value_hash)
        .join(CallNode, Argument.call_hash == CallNode.call_hash)
        .join(Job, CallNode.call_hash == Job.call_hash)
        .filter(File.path == path)
        .order_by(Job.start_time.desc())
        .first()
    )

    # Search for File as a Argument to a Task.
    row4 = (
        session.query(File, Job)
        .join(Subvalue, File.value_hash == Subvalue.value_hash)
        .join(Argument, Subvalue.parent_value_hash == Argument.value_hash)
        .join(CallNode, Argument.call_hash == CallNode.call_hash)
        .join(Job, CallNode.call_hash == Job.call_hash)
        .filter(File.path == path)
        .order_by(Job.start_time.desc())
        .first()
    )

    if row3 or row4:
        # Return the most recent file and job reference.
        if not row3:
            return (row4[0], row4[1], "arg")
        elif not row4:
            return (row3[0], row3[1], "arg")
        else:
            _, job3 = row3
            _, job4 = row4
            if job3.end_time > job4.end_time:
                return (row3[0], row3[1], "arg")
            else:
                return (row4[0], row4[1], "arg")

    return None



"""change timestamps to utc

Revision ID: 3b0a6e67cc58
Revises: f68b3aaee9cc
Create Date: 2024-08-24 14:20:17.063050

"""

import os

from alembic import op


# revision identifiers, used by Alembic.
revision = "3b0a6e67cc58"
down_revision = "f68b3aaee9cc"
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == "postgresql":
        # Users can define what local timezone should be assumed for their
        # database using this environment variable.
        local_timezone = os.environ.get("REDUN_LOCALTIMEZONE", "UTC")

        op.execute(
            f"""
            alter table job
            alter column start_time type timestamp with time zone;

            alter table job
            alter column end_time type timestamp with time zone;

            alter table call_node
            alter column "timestamp" type timestamp with time zone;

            alter table redun_version
            alter column "timestamp" type timestamp with time zone;

            -- Create trigger that only runs for old clients to help convert their timestamps.
            CREATE OR REPLACE FUNCTION check_job_utc()
            RETURNS TRIGGER AS $$
            BEGIN
                IF current_setting('redun.version', true) IS NULL THEN
                    -- For older redun clients, convert time from local timezone to UTC.
                    set local timezone = '{local_timezone}';
                    NEW.start_time := NEW.start_time AT TIME ZONE 'UTC';
                    NEW.end_time := NEW.end_time AT TIME ZONE 'UTC';
                END IF;

                RETURN NEW;
            END;
            $$ LANGUAGE plpgsql;


            DROP TRIGGER IF EXISTS check_job_utc ON job;
            CREATE TRIGGER check_job_utc
            BEFORE INSERT ON job
            FOR EACH ROW
            EXECUTE FUNCTION check_job_utc();
            """
        )

    elif op.get_bind().dialect.name == "sqlite":
        op.execute(
            """
            update job set
              start_time = datetime(start_time, 'utc'),
              end_time = datetime(end_time, 'utc');
            """
        )


def downgrade():
    if op.get_bind().dialect.name == "postgresql":
        op.execute(
            """
            alter table job
            alter column start_time type timestamp without time zone;

            alter table job
            alter column end_time type timestamp without time zone;

            alter table call_node
            alter column "timestamp" type timestamp without time zone;

            alter table redun_version
            alter column "timestamp" type timestamp without time zone;

            DROP TRIGGER IF EXISTS check_job_utc ON job;
            DROP FUNCTION IF EXISTS check_job_utc;
            """
        )

    elif op.get_bind().dialect.name == "sqlite":
        op.execute(
            """
            update job set
              start_time = datetime(start_time, 'localtime'),
              end_time = datetime(end_time, 'localtime');
            """
        )



import json
import re
from typing import Any, Tuple, Union

from redun.utils import trim_string

# Built-in redun tag keys.
VERSION_KEY = "redun.version"
PROJECT_KEY = "project"
DOC_KEY = "doc"
USER_KEY = "user"
CONTEXT_KEY = "redun.context"

# Matches any value.
ANY_VALUE = object()


def str2literal(value_str: str) -> Union[bool, None]:
    """
    Parse a JSON literal.
    """
    lookup = {
        "true": True,
        "false": False,
        "null": None,
    }
    if value_str in lookup:
        return lookup[value_str]
    else:
        raise ValueError(f"Unknown JSON literal: {value_str}")


def parse_tag_value(value_str: str) -> Any:
    """
    Parse a tag value from the cli arguments.
    """
    # Empty string is short for None.
    if not value_str:
        return None

    # If the first character is a JSON compound type (array, object) or
    # or a string, then parse as normal json.
    if value_str[0] in ("[", "{", '"'):
        try:
            return json.loads(value_str)
        except json.JSONDecodeError as error:
            raise ValueError(f"Invalid tag value: {error}")

    # Try to automatically infer the type of the tag value.
    try:
        return int(value_str)
    except ValueError:
        pass

    try:
        return float(value_str)
    except ValueError:
        pass

    try:
        return str2literal(value_str)
    except ValueError:
        pass

    # Assume string.
    return value_str


def parse_tag_key_value(key_value: str, value_required=True) -> Tuple[str, Any]:
    """
    Parse a tag key=value pair from cli arguments.
    """
    if not key_value:
        raise ValueError("key must be specified.")

    if "=" not in key_value:
        if value_required:
            raise ValueError(f"key=value pair expected: '{key_value}'")
        return (key_value, ANY_VALUE)

    key, value = key_value.split("=", 1)
    if not key:
        raise ValueError(f"key must be specified: '{key_value}'")
    return (key, parse_tag_value(value))


def format_tag_value(value: Any) -> str:
    """
    Format a tag value.
    """
    # Simple strings (no spaces or commas or special values) can be displayed without quotes.
    if (
        isinstance(value, str)
        and not re.match(".*[ ,].*", value)
        and isinstance(parse_tag_value(value), str)
    ):
        return value
    else:
        return json.dumps(value, sort_keys=True)


def format_tag_key_value(key: str, value: Any, max_length: int = 50) -> str:
    """
    Format a tag key value pair.
    """
    key = trim_string(key, max_length=max_length)
    value_str = trim_string(format_tag_value(value), max_length=max_length)
    return f"{key}={value_str}"



import os
import threading
import time
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, cast

from redun.scheduler import Job

# Per AWS documentation, see:
# https://docs.aws.amazon.com/batch/latest/userguide/service_limits.html
MAX_ARRAY_SIZE = 10000
# https://docs.aws.amazon.com/batch/latest/userguide/job_env_vars.html
AWS_ARRAY_VAR = "AWS_BATCH_JOB_ARRAY_INDEX"
# https://kubernetes.io/docs/tasks/job/indexed-parallel-processing-static/
K8S_ARRAY_VAR = "JOB_COMPLETION_INDEX"
# https://cloud.google.com/batch/docs/create-run-basic-job#predefined-environment-variables
GCP_ARRAY_VAR = "BATCH_TASK_INDEX"


def get_job_array_index(env: dict = cast(dict, os.environ)) -> Optional[int]:
    """
    Finds any Job Array Index environment variables and returns the index.
    """
    if AWS_ARRAY_VAR in env:
        return int(env[AWS_ARRAY_VAR])
    elif K8S_ARRAY_VAR in env:
        return int(env[K8S_ARRAY_VAR])
    elif GCP_ARRAY_VAR in env:
        return int(env[GCP_ARRAY_VAR])
    else:
        return None


class JobDescription:
    def __init__(self, job: Job):
        self.task_name = job.task.fullname
        self.options = job.get_options()
        self.key = self.task_name + " " + str(sorted(self.options.items()))

    def __hash__(self) -> int:
        return hash(self.key)

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, JobDescription) and self.key == other.key

    def __repr__(self) -> str:
        return f"NAME:{self.task_name}\tOPTIONS: {self.options}"


class JobArrayer:
    """
    A 'sink' for submitted jobs that detects when jobs can be submitted as an
    array job. Eligible jobs will have the same task, task options, and
    resource requirements, but can differ in the args and kwargs passed to the
    task.

    The method uses "staleness" of submitted jobs in the 'sink'.  Eligible jobs
    are grouped in the pool (really a dict). Grouped jobs that haven't had new
    ones added to the group for `stale_time` sec will be submitted.

    Parameters
    ----------
    submit_jobs: Callable[[List[Job]], None]
        Callback called when a group of jobs are ready to be submitted as an array.

    on_error: Callable[[Exception], None]
        Callback called when there is a top-level error.

    submit_interval: float
        How frequently the monitor thread will check for submittable stale
        jobs. Should be less than `stale_time`, ideally.

    stale_time: float
        Job groupings that haven't had new jobs added for this many seconds
        will be submitted.

    min_array_size: int
        Minimum number of jobs in a group to be submitted as an array job
        instead of individual jobs. Can be anywhere from 2 to
        `MAX_ARRAY_SIZE-1`.

    max_array_size: int
        Maximum number of jobs that can be submitted as an array job. Must
        be in (`min_array_size`, `MAX_ARRAY_SIZE`].
    """

    def __init__(
        self,
        submit_jobs: Callable[[List[Job]], None],
        on_error: Callable[[Exception], None],
        submit_interval: float,
        stale_time: float,
        min_array_size: int,
        max_array_size: int = MAX_ARRAY_SIZE,
    ):
        self.min_array_size = min_array_size
        self.max_array_size = min(max_array_size, MAX_ARRAY_SIZE)
        if self.max_array_size < self.min_array_size:
            raise ValueError("Maximum array size cannot be less than minimum.")

        self.pending: Dict[JobDescription, List[Job]] = defaultdict(list)
        self.pending_timestamps: Dict[JobDescription, float] = {}
        self._lock = threading.Lock()

        self._submit_jobs = submit_jobs
        self._on_error = on_error
        self.interval = submit_interval
        self.stale_time = stale_time

        # Monitor thread
        self._monitor_thread = threading.Thread(target=self._monitor_stale_jobs, daemon=True)
        self._exit_flag = threading.Event()
        self.num_pending = 0

    def _monitor_stale_jobs(self) -> None:
        """Monitoring thread batches up stale jobs"""
        try:
            while not self._exit_flag.wait(timeout=self.interval):
                stales = self.get_stale_descrs()

                for descr in stales:
                    self.submit_pending_jobs(descr)
        except Exception as error:
            # Since we run this method at the top level of a thread, we need to
            # catch all exceptions so we can properly report them to the
            # scheduler.
            self._on_error(error)

    def start(self) -> None:
        # Do not have a monitor thread when arraying is disabled.
        if not self.min_array_size:
            return

        if self._monitor_thread.is_alive():
            return

        # Initialize a new Thread here in case a previous one has completed,
        # since Threads can't be started more than once.
        self._exit_flag.clear()
        self._monitor_thread = threading.Thread(target=self._monitor_stale_jobs, daemon=True)
        self._monitor_thread.start()

    def stop(self) -> None:
        self._exit_flag.set()
        if self._monitor_thread.is_alive():
            self._monitor_thread.join()

    def add_job(self, job: Job) -> None:
        """Adds a new job"""

        # If arraying is turned off, just submit the job.
        # Script jobs are also not handled yet.
        if job.task.script or not self.min_array_size:
            self._submit_jobs([job])
            return

        descr = JobDescription(job)
        with self._lock:
            self.pending[descr].append(job)
            self.pending_timestamps[descr] = time.time()
            self.num_pending += 1

        self.start()

    def get_stale_descrs(self) -> List[JobDescription]:
        """Submits jobs that haven't been touched in a while"""
        currtime = time.time()
        stales = [
            descr
            for descr in self.pending
            if (currtime - self.pending_timestamps[descr] > self.stale_time)
        ]
        return stales

    def submit_pending_jobs(self, descr: JobDescription) -> None:
        # Lock, otherwise adding a job to this descr at the wrong time could
        # result in timestamps out of sync with jobs.
        with self._lock:
            jobs = self.pending.pop(descr)
            timestamp = self.pending_timestamps.pop(descr)

        # Submit at most max_array_size jobs and leave the remainder.
        if len(jobs) > self.max_array_size:
            remainder = jobs[self.max_array_size :]
            jobs = jobs[: self.max_array_size]
            self._submit_jobs(jobs)

            with self._lock:
                self.pending[descr].extend(remainder)
                self.pending_timestamps[descr] = timestamp

        elif len(jobs) < self.min_array_size:
            for job in jobs:
                self._submit_jobs([job])
        else:
            self._submit_jobs(jobs)

        self.num_pending -= len(jobs)
